import os
import sys
import numpy


from amuse.test.amusetest import TestWithMPI
from amuse.community.ramses.interface import RamsesInterface
from amuse.community.ramses.interface import Ramses
from amuse.units import generic_unit_system
from amuse.units import generic_unit_converter
from amuse.units import units
from amuse import datamodel

class TestRamsesInterface(TestWithMPI):
    
    def test0(self):
        instance=RamsesInterface(redirection="none")
        instance.set_input_directory(instance.get_default_input_directory())
        instance.initialize_code()
        instance.commit_parameters()
        instance.stop()
        
    def test1(self):
        instance=RamsesInterface(mode="1d", redirection="none")
        instance.set_input_directory(instance.get_default_input_directory())
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.commit_parameters())
        self.assertEqual(0, instance.setup_mesh(10, 1, 1, 1.0, 0.0, 0.0))
        self.assertEqual(0, instance.initialize_grid())
        nx, ny, nz, error = instance.get_mesh_size()
        self.assertEqual(0, error)
        self.assertEqual((nx, ny, nz), (10, 1, 1))
        self.assertEqual(0, instance.evolve_model(0.1))
        time, error = instance.get_time()
        self.assertEqual(0, error)
        self.assertAlmostEqual(time, 0.1, 2)
        instance.stop()
    
    def xtest1b(self):
        instance=RamsesInterface()
        instance.initialize_code()
        instance.setup_mesh(50,40,30,1.,1.,1.)
        nx,ny,nz, error = instance.get_mesh_size()
        self.assertEqual(nx, 50)
        self.assertEqual(ny, 40)
        self.assertEqual(nz, 30)
        instance.stop()
    
    def xtest2(self):
        instance=RamsesInterface()
        instance.initialize_code()
        instance.setup_mesh(50,50,50,1.,1.,1.)
        instance.commit_parameters()
        instance.set_grid_state(1,1,1,1.,0.1,0.1,0.1,1.)
        rho, err=instance.get_grid_density(1,1,1)
        self.assertEqual(rho,1.)
        rhovx,rhovy,rhovz,err=instance.get_grid_momentum_density(1,1,1)
        self.assertEqual(rhovx,0.1)
        self.assertEqual(rhovy,0.1)
        self.assertEqual(rhovz,0.1)
        en,err=instance.get_grid_energy_density(1,1,1)
        self.assertEqual(en,1.)
        rho,rhovx,rhovy,rhovz,en,err=instance.get_grid_state(1,1,1)
        self.assertEqual(rho,1.)
        self.assertEqual(rhovx,0.1)
        self.assertEqual(rhovy,0.1)
        self.assertEqual(rhovz,0.1)
        self.assertEqual(en,1.)
        instance.stop()
    
    def xtest3(self):
        instance=RamsesInterface()
        instance.initialize_code()
        instance.setup_mesh(50,50,50,1.,1.,1.)
        instance.commit_parameters()
        x,y,z=numpy.indices( (50,50,50) )
        x=x.flatten()+1
        y=y.flatten()+1
        z=z.flatten()+1
        rho=0.25*numpy.ones_like(x)
        rhvx=0.*numpy.ones_like(x)
        rhvy=0.*numpy.ones_like(x)
        rhvz=0.*numpy.ones_like(x)
        en=2.*numpy.ones_like(x)
        instance.set_grid_state(x,y,z,rho,rhvx,rhvy,rhvz,en)
        rho,rhovx,rhovy,rhovz,en,err=instance.get_grid_state(1,1,1)
        self.assertEqual(rho,0.25)
        self.assertEqual(rhovx,0.)
        self.assertEqual(rhovy,0.)
        self.assertEqual(rhovz,0.)
        self.assertEqual(en,2.)
        instance.stop()
    
    def xtest4(self):
        instance=RamsesInterface()
        instance.initialize_code()
        instance.setup_mesh(50,50,50,1.,1.,1.)
        instance.commit_parameters()
        x,y,z=numpy.indices( (50,50,50) )
        x=x.flatten()+1
        y=y.flatten()+1
        z=z.flatten()+1
        rho=0.1*numpy.ones_like(x)
        rhvx=0.*numpy.ones_like(x)
        rhvy=0.*numpy.ones_like(x)
        rhvz=0.*numpy.ones_like(x)
        en=0.1*numpy.ones_like(x)
        instance.set_grid_state(x,y,z,rho,rhvx,rhvy,rhvz,en)
        instance.initialize_grid()
        instance.stop()
    
    def xtest5(self):
        instance=RamsesInterface()
        instance.initialize_code()
        instance.setup_mesh(40,40,40,1.,1.,1.)
        instance.commit_parameters()
        x,y,z=numpy.indices( (40,40,40) )
        x=x.flatten()+1
        y=y.flatten()+1
        z=z.flatten()+1
        rho=0.1*numpy.ones_like(x)
        rhvx=0.*numpy.ones_like(x)
        rhvy=0.*numpy.ones_like(x)
        rhvz=0.*numpy.ones_like(x)
        en=0.1*numpy.ones_like(x)
        instance.set_grid_state(x,y,z,rho,rhvx,rhvy,rhvz,en)
        instance.initialize_grid()
        instance.evolve_model(0.01)
        tnow,err=instance.get_time()
        self.assertAlmostEqual(tnow,0.01,15)
        instance.evolve_model(0.025)
        tnow,err=instance.get_time()
        self.assertAlmostEqual(tnow,0.025,15)
        instance.evolve_model(0.025001)
        tnow,err=instance.get_time()
        self.assertAlmostEqual(tnow,0.025001,15)
        instance.evolve_model(0.0321)
        tnow,err=instance.get_time()
        self.assertAlmostEqual(tnow,0.0321,15)
        instance.evolve_model(0.0321)
        tnow,err=instance.get_time()
        self.assertAlmostEqual(tnow,0.0321,15)
        instance.evolve_model(0.07)
        tnow,err=instance.get_time()
        self.assertAlmostEqual(tnow,0.07,15)
        instance.stop()
    
    def xtest6(self):
        instance=RamsesInterface(number_of_workers=1)
        instance.initialize_code()
        instance.setup_mesh(150,10,30,1.,1.,1.)
        instance.commit_parameters()
        x,y,z=numpy.indices( (150,10,30) )
        x=x.flatten()+1
        y=y.flatten()+1
        z=z.flatten()+1
        rho=0.1*numpy.ones_like(x)
        rhvx=0.*numpy.ones_like(x)
        rhvy=0.*numpy.ones_like(x)
        rhvz=0.*numpy.ones_like(x)
        en=0.1*numpy.ones_like(x)
        instance.set_grid_state(x,y,z,rho,rhvx,rhvy,rhvz,en)
        instance.initialize_grid()
        instance.evolve_model(0.01)
        x,y,z,err=instance.get_position_of_index(15,5,20)
        self.assertAlmostEqual(x,15/150.-1/300.,15)
        self.assertAlmostEqual(y,5/10.-1/20.,15)
        self.assertAlmostEqual(z,20/30.-1/60.,15)
        i,j,k,err=instance.get_index_of_position(x,y,z)
        self.assertEqual([i,j,k],[15,5,20])
        instance.stop()

    def xtest7(self):
        instance=RamsesInterface()
        instance.initialize_code()
        instance.setup_mesh(50,50,50,1.,1.,1.)
        instance.commit_parameters()
        err=instance.set_gravity_field(1,2,3,1.,0.5,0.25)
        self.assertEqual(err,0)
        fx,fy,fz,err=instance.get_gravity_field(1,2,3)
        self.assertEqual(fx,1.)
        self.assertEqual(fy,0.5)
        self.assertEqual(fz,0.25)
        instance.stop()

    def xtest8(self):
        instance=RamsesInterface()
        instance.initialize_code()
        err=instance.set_boundary("periodic","reflective",
        "periodic","reflective",
        "periodic","reflective")
        self.assertEqual(err,-1)
        instance.stop()
    
        instance=RamsesInterface()
        instance.initialize_code()
        err=instance.set_boundary("reflective","periodic",
        "periodic","reflective",
        "periodic","reflective")
        self.assertEqual(err,-2)
        instance.stop()
    
        instance=RamsesInterface()
        instance.initialize_code()
        err=instance.set_boundary("periodic","periodic",
        "periodic","periodic",
        "periodic","periodic")
        self.assertEqual(err,0)
        instance.stop()

    def xtest9(self):
        instance=RamsesInterface(number_of_workers=2)
        instance.initialize_code()
        instance.setup_mesh(50,50,50,1.,1.,1.)
        instance.commit_parameters()
        instance.set_grid_state(1,1,1,1.,0.1,0.1,0.1,1.)
        instance.set_grid_state(50,50,50,2.,0.2,0.2,0.2,2.)

        rho, err=instance.get_grid_density(1,1,1)
        self.assertEqual(rho,1.)
        rhovx,rhovy,rhovz,err=instance.get_grid_momentum_density(1,1,1)
        self.assertEqual(rhovx,0.1)
        self.assertEqual(rhovy,0.1)
        self.assertEqual(rhovz,0.1)
        en,err=instance.get_grid_energy_density(1,1,1)
        self.assertEqual(en,1.)
        rho,rhovx,rhovy,rhovz,en,err=instance.get_grid_state(1,1,1)
        self.assertEqual(rho,1.)
        self.assertEqual(rhovx,0.1)
        self.assertEqual(rhovy,0.1)
        self.assertEqual(rhovz,0.1)
        self.assertEqual(en,1.)

        rho, err=instance.get_grid_density(50,50,50)
        self.assertEqual(err,0)
        self.assertEqual(rho,2.)
        rhovx,rhovy,rhovz,err=instance.get_grid_momentum_density(50,50,50)
        self.assertEqual(rhovx,0.2)
        self.assertEqual(rhovy,0.2)
        self.assertEqual(rhovz,0.2)
        en,err=instance.get_grid_energy_density(50,50,50)
        self.assertEqual(en,2.)
        rho,rhovx,rhovy,rhovz,en,err=instance.get_grid_state(50,50,50)
        self.assertEqual(rho,2.)
        self.assertEqual(rhovx,0.2)
        self.assertEqual(rhovy,0.2)
        self.assertEqual(rhovz,0.2)
        self.assertEqual(en,2.)


        instance.stop()
        
    def xtest10(self):
        instance=self.new_instance(RamsesInterface)
        instance.initialize_code()
        instance.setup_mesh(100,5,5,100.0,0,0)
        instance.set_boundary("interface","periodic","periodic","periodic","periodic","periodic")
        instance.commit_parameters()
        
        minx, maxx, miny, maxy, minz, maxz, error = instance.get_boundary_index_range_inclusive(1)
        self.assertEqual(error, 0)
        self.assertEqual(minx, 1)
        self.assertEqual(maxx, 2)
        self.assertEqual(miny, 1)
        self.assertEqual(maxy, 5)
        self.assertEqual(minz, 1)
        self.assertEqual(maxz, 5)
        
        for i in range(2,7):
            minx, maxx, miny, maxy, minz, maxz, error = instance.get_boundary_index_range_inclusive(i)
            self.assertEqual(error, 0)
            self.assertEqual(minx, 1)
            self.assertEqual(maxx, 1)
            self.assertEqual(miny, 1)
            self.assertEqual(maxy, 1)
            self.assertEqual(minz, 1)
            self.assertEqual(maxz, 1)
    
    def xtest11(self):
        instance=self.new_instance(RamsesInterface)
        instance.initialize_code()
        instance.setup_mesh(100,5,6,100.0,0,0)
        instance.set_boundary("interface","interface","interface","interface","interface","interface")
        instance.commit_parameters()
        
        
        for i in range(1,7):
            minx, maxx, miny, maxy, minz, maxz, error = instance.get_boundary_index_range_inclusive(i)
            self.assertEqual(error, 0),
            self.assertEqual(minx, 1)
            self.assertEqual(miny, 1)
            self.assertEqual(minz, 1)
            if i == 1 or i == 2:
                self.assertEqual(maxx, 2)
                self.assertEqual(maxy, 5)
                self.assertEqual(maxz, 6)
            elif i == 3 or i == 4:
                self.assertEqual(maxx, 100+4)
                self.assertEqual(maxy, 2)
                self.assertEqual(maxz, 6)
            elif i == 5 or i == 6:
                self.assertEqual(maxx, 100+4)
                self.assertEqual(maxy, 5+4)
                self.assertEqual(maxz, 2)
    
    def xtest12(self):
        instance=self.new_instance(RamsesInterface)
        instance.initialize_code()
        instance.setup_mesh(100,2,2,100.0,100.0,100.0)
        instance.set_boundary("interface","periodic","periodic","periodic","periodic","periodic")
        instance.commit_parameters()
        
        for i in [1,2]:
            error = instance.set_boundary_state(
                i,1,1,       #  index
                1.0 * (i+1),         #  density
                2.0 * (i+1), 3.0 * (i+1), 4.0 * (i+1), #  momentum
                5.0 * (i+1),         #  energy
                1                    #  boundary
            )
            self.assertEqual(error, 0)
            rho, rhovx, rhovy, rhovz, rhoen, error = instance.get_boundary_state(
                i, 1, 1,
                1
            )
            print(rho, rhovx, rhovy, rhovz, rhoen, error) 
            self.assertEqual(error, 0)
            self.assertAlmostRelativeEquals(rho, 1.0 * (i+1))
            self.assertAlmostRelativeEquals(rhovx, 2.0 * (i+1))
            self.assertAlmostRelativeEquals(rhovy, 3.0 * (i+1))
            self.assertAlmostRelativeEquals(rhovz, 4.0 * (i+1))
            self.assertAlmostRelativeEquals(rhoen, 5.0 * (i+1))
    
    def xtest13(self):
        instance=self.new_instance(RamsesInterface)
        instance.initialize_code()
        instance.setup_mesh(100,2,2,100.0,0,0)
        instance.set_boundary("interface","interface","periodic","periodic","periodic","periodic")
        instance.commit_parameters()
        
        for i in [1,2]:
            for j in [1,2]:
                error = instance.set_boundary_state(
                    i,1,1,       #  index
                    1.0 * (i+1),         #  density
                    2.0 * (i+1), 3.0 * (i+1), 4.0 * (i+1), #  momentum
                    5.0 * (i+1),         #  energy
                    j    #  boundary 
                )
                self.assertEqual(error, 0)
                rho, rhovx, rhovy, rhovz, rhoen, error = instance.get_boundary_state(
                    i, 1,1,
                    j
                )
                print(j)
                self.assertEqual(error, 0)
                
                self.assertAlmostRelativeEquals(rho, 1.0 * (i+1))
                self.assertAlmostRelativeEquals(rhovx, 2.0 * (i+1))
                self.assertAlmostRelativeEquals(rhovy, 3.0 * (i+1))
                self.assertAlmostRelativeEquals(rhovz, 4.0 * (i+1))
                self.assertAlmostRelativeEquals(rhoen, 5.0 * (i+1))
    
    def xtest14(self):
        instance=self.new_instance(RamsesInterface)
        instance.initialize_code()
        instance.setup_mesh(5,6,7,100.0,100.0,100.0)
        instance.set_boundary("interface","interface","interface","interface","interface","interface")
        instance.commit_parameters()
        
        x1range = (2,6,7)
        x2range = (5+4,2,7)
        x3range = (5+4,6+4,2)
    
        for xrange, j in zip([x1range, x1range, x2range, x2range, x3range, x3range], [1,2,3,4,5,6]):
            for i0 in range(xrange[0]):
                for j0 in range(xrange[1]):
                    for k0 in range(xrange[2]):
                        i = (i0 * (xrange[2] * xrange[1])) + (j0 * xrange[2]) + k0
                        print("boundary:", j, i0+1, j0+1, k0+1)
                        error = instance.set_boundary_state(
                            i0+1, j0+1, k0+1,       #  index
                            1.0 * (i+1),         #  density
                            2.0 * (i+1), 3.0 * (i+1), 4.0 * (i+1), #  momentum
                            5.0 * (i+1),         #  energy
                            j
                        )
                        self.assertEqual(error, 0)
                        rho, rhovx, rhovy, rhovz, rhoen, error = instance.get_boundary_state(
                            i0+1, j0+1, k0+1,       #  index
                            j
                        )
                        self.assertEqual(error, 0)
                        
                        self.assertAlmostRelativeEquals(rho, 1.0 * (i+1))
                        self.assertAlmostRelativeEquals(rhovx, 2.0 * (i+1))
                        self.assertAlmostRelativeEquals(rhovy, 3.0 * (i+1))
                        self.assertAlmostRelativeEquals(rhovz, 4.0 * (i+1))
                        self.assertAlmostRelativeEquals(rhoen, 5.0 * (i+1))
    
    def xtest15(self):
        instance=self.new_instance(RamsesInterface)
        instance.initialize_code()
        instance.setup_mesh(100,5,4,100.0,100.0, 100.0)
        instance.set_boundary("interface","interface","periodic","periodic","periodic","periodic")
        instance.commit_parameters()
        
        dx = 100.0 / 100.0
        dy = 100.0 / 5.0
        dz = 100.0 / 4.0
        
        for i in [1,2]:
            x,y,z,error = instance.get_boundary_position_of_index(
                i, 1, 1,
                1
            )
            self.assertEqual(error, 0)
            self.assertAlmostRelativeEquals(x, (0.5 * dx) - (i * dx))
            self.assertAlmostRelativeEquals(y, (0.5 * dy))
            self.assertAlmostRelativeEquals(z, (0.5 * dz))
    
    def xtest16(self):
        instance=self.new_instance(RamsesInterface)
        instance.initialize_code()
        instance.setup_mesh(100,5,4,100.0,100.0, 100.0)
        instance.set_boundary("interface","interface","periodic","periodic","periodic","periodic")
        instance.commit_parameters()
        
        dx = 100.0 / 100.0
        dy = 100.0 / 5.0
        dz = 100.0 / 4.0
        
        for i in [1,2]:
            x,y,z,error = instance.get_boundary_position_of_index(
                i,1,1,
                2
            )
            self.assertEqual(error, 0)
            self.assertAlmostRelativeEquals(x, 100.0 + (0.5 * dx) + ((i-1) * dx))
            self.assertAlmostRelativeEquals(y, (0.5 * dy))
            self.assertAlmostRelativeEquals(z, (0.5 * dz))
    
    def xtest17(self):
        instance=self.new_instance(RamsesInterface)
        instance.initialize_code()
        instance.setup_mesh(100,5,4,100.0,100.0, 100.0)
        instance.set_boundary("interface","interface","interface","interface","periodic","periodic")
        instance.commit_parameters()
        
        dx = 100.0 / 100.0
        dy = 100.0 / 5.0
        dz = 100.0 / 4.0
        
        for i in [1,2]:
            for j in range(1,6):
                x,y,z,error = instance.get_boundary_position_of_index(
                    i, j, 1, 
                    2
                )
                self.assertEqual(error, 0)
                self.assertAlmostRelativeEquals(x, 100.0 + (0.5 * dx) + ((i-1) * dx))
                self.assertAlmostRelativeEquals(y, (0.5 * dy) + ((j-1) * dy))
                self.assertAlmostRelativeEquals(z, (0.5 * dz))
        
        for i in range(1, 100 + 4 + 1):
            for j in [1,2]:
                x,y,z,error = instance.get_boundary_position_of_index(
                    i, j, 1, 
                    3
                )
                self.assertEqual(error, 0)
                self.assertAlmostRelativeEquals(x, (0.5 * dx) + ((i-2-1) * dx))
                self.assertAlmostRelativeEquals(y, 0.0 - ((0.5 * dy) + ((j-1) * dy)))
                self.assertAlmostRelativeEquals(z, (0.5 * dz))
                
                
                x,y,z,error = instance.get_boundary_position_of_index(
                    i, j, 1, 
                    4
                )
                self.assertEqual(error, 0)
                self.assertAlmostRelativeEquals(x, (0.5 * dx) + ((i-2-1) * dx))
                self.assertAlmostRelativeEquals(y, 100.0 + (0.5 * dy) + ((j-1) * dy))
                self.assertAlmostRelativeEquals(z, (0.5 * dz))
    
    def xtest18(self):
        instance=self.new_instance(RamsesInterface)
        instance.initialize_code()
        instance.setup_mesh(3, 3, 3, 6,12,18)
        instance.set_boundary("interface","interface","interface","interface","interface","interface")
        instance.commit_parameters()
        
        dx = 6.0 / 3.0
        dy = 12.0 / 3.0
        dz = 18.0 / 3.0
        for i in [1,2]:
            for j in range(1,3+1):
                for k in range(1,3+1):
                    x,y,z,error = instance.get_boundary_position_of_index(
                        i, j, k, 
                        2
                    )
                    self.assertEqual(error, 0)
                    self.assertAlmostRelativeEquals(x, 6.0 + (0.5 * dx) + ((i-1) * dx))
                    self.assertAlmostRelativeEquals(y, (0.5 * dy) + ((j-1) * dy))
                    self.assertAlmostRelativeEquals(z, (0.5 * dz) + ((k-1) * dz))
        
        for i in range(1,3 + 4 +1):
            for j in [1,2]:
                for k in range(1,3+1):
                    x,y,z,error = instance.get_boundary_position_of_index(
                        i, j, k, 
                        3
                    )
                    self.assertEqual(error, 0)
                    self.assertAlmostRelativeEquals(x, (0.5 * dx) + ((i-2-1) * dx))
                    self.assertAlmostRelativeEquals(y, 0.0 - ((0.5 * dy) + ((j-1) * dy)))
                    self.assertAlmostRelativeEquals(z, (0.5 * dz) + ((k-1) * dz))
                    
                    
                    x,y,z,error = instance.get_boundary_position_of_index(
                        i, j, k, 
                        4
                    )
                    self.assertEqual(error, 0)
                    self.assertAlmostRelativeEquals(x, (0.5 * dx) + ((i-2-1) * dx))
                    self.assertAlmostRelativeEquals(y, 12.0 + (0.5 * dy) + ((j-1) * dy))
                    self.assertAlmostRelativeEquals(z, (0.5 * dz) + ((k-1) * dz))
        
        for i in range(1,3 + 4 +1):
            for j in range(1,3 + 4 +1):
                for k in [1,2]:
                    x,y,z,error = instance.get_boundary_position_of_index(
                        i, j, k, 
                        5 
                    )
                    self.assertEqual(error, 0)
                    self.assertAlmostRelativeEquals(x, (0.5 * dx) + ((i-2-1) * dx))
                    self.assertAlmostRelativeEquals(y, (0.5 * dy) + ((j-2-1) * dy))
                    self.assertAlmostRelativeEquals(z,  0.0 - ((0.5 * dz) + ((k-1) * dz)))
                    
                    
                    x,y,z,error = instance.get_boundary_position_of_index(
                        i, j, k, 
                        6
                    )
                    self.assertEqual(error, 0)
                    self.assertAlmostRelativeEquals(x, (0.5 * dx) + ((i-2-1) * dx))
                    self.assertAlmostRelativeEquals(y, (0.5 * dy) + ((j-2-1) * dy))
                    self.assertAlmostRelativeEquals(z, 18.0 + (0.5 * dz) + ((k-1) * dz))
        
    def xtest19(self):
        results = []
        instance=self.new_instance(RamsesInterface)
        instance.initialize_code()
        instance.commit_parameters()
        nx, ny, nz, error = instance.get_parallel_decomposition()
        self.assertEqual(error, 0)
        self.assertEqual(nx, 1)
        self.assertEqual(ny, 1)
        self.assertEqual(nz, 1)
        error = instance.set_parallel_decomposition(2,1,1)
        self.assertEqual(error, -1)
        
   
    def xtest20(self):
        results = []
        instance=self.new_instance(RamsesInterface, number_of_workers = 4)
        instance.initialize_code()
        nx, ny, nz, error = instance.get_parallel_decomposition()
        self.assertEqual(error, 0)
        self.assertEqual(nx, 0)
        self.assertEqual(ny, 0)
        self.assertEqual(nz, 0)
        error = instance.set_parallel_decomposition(2,1,2)
        self.assertEqual(error, 0)
        nx, ny, nz, error = instance.get_parallel_decomposition()
        self.assertEqual(error, 0)
        self.assertEqual(nx, 2)
        self.assertEqual(ny, 1)
        self.assertEqual(nz, 2)
        error = instance.set_parallel_decomposition(10,3,2)
        self.assertEqual(error, -1)
        
    def xtest21(self):
        results = []
        instance=self.new_instance(RamsesInterface, number_of_workers = 2)
        instance.initialize_code()
        error = instance.set_parallel_decomposition(1,2,1)
        self.assertEqual(error, 0)
        instance.setup_mesh(10,30,10,100.0, 300.0, 100.0)
        instance.set_boundary("interface","interface","periodic","periodic","periodic","periodic")
        instance.commit_parameters()
        
           
        for boundary_index in [1,2]:
            for i0 in range(1,2):
                for j0 in range(1, 30+1):
                    i = j0 * 30 + i0
                    error = instance.set_boundary_state(
                        i0, j0, 1,       #  index
                        1.0 * (i+1),         #  density
                        2.0 * (i+1), 3.0 * (i+1), 4.0 * (i+1), #  momentum
                        5.0 * (i+1),         #  energy
                        boundary_index     #  boundary 
                    )
                    self.assertEqual(error, 0)
                    rho, rhovx, rhovy, rhovz, rhoen, error = instance.get_boundary_state(
                        i0, j0, 1,
                        boundary_index
                    )
                    self.assertEqual(error, 0)
                    
                    self.assertAlmostRelativeEquals(rho, 1.0 * (i+1))
                    self.assertAlmostRelativeEquals(rhovx, 2.0 * (i+1))
                    self.assertAlmostRelativeEquals(rhovy, 3.0 * (i+1))
                    self.assertAlmostRelativeEquals(rhovz, 4.0 * (i+1))
                    self.assertAlmostRelativeEquals(rhoen, 5.0 * (i+1))
    
    def xtest22(self):
        results = []
        instance=self.new_instance(RamsesInterface, number_of_workers = 2)
        instance.initialize_code()
        error = instance.set_parallel_decomposition(2,1,1)
        self.assertEqual(error, 0)
        instance.setup_mesh(10,30,10,100.0, 300.0, 100.0)
        instance.set_boundary("interface","interface","periodic","periodic","periodic","periodic")
        instance.commit_parameters()
        
           
        for boundary_index in [1,2]:
            for i0 in range(1,2):
                for j0 in range(1, 30+1):
                    i = j0 * 30 + i0
                    error = instance.set_boundary_state(
                        i0, j0, 1,       #  index
                        1.0 * (i+1),         #  density
                        2.0 * (i+1), 3.0 * (i+1), 4.0 * (i+1), #  momentum
                        5.0 * (i+1),         #  energy
                        boundary_index     #  boundary 
                    )
                    self.assertEqual(error, 0)
                    rho, rhovx, rhovy, rhovz, rhoen, error = instance.get_boundary_state(
                        i0, j0, 1,
                        boundary_index
                    )
                    self.assertEqual(error, 0)
                    
                    self.assertAlmostRelativeEquals(rho, 1.0 * (i+1))
                    self.assertAlmostRelativeEquals(rhovx, 2.0 * (i+1))
                    self.assertAlmostRelativeEquals(rhovy, 3.0 * (i+1))
                    self.assertAlmostRelativeEquals(rhovz, 4.0 * (i+1))
                    self.assertAlmostRelativeEquals(rhoen, 5.0 * (i+1))
    
    def xtest23(self):
        results = []
        instance=self.new_instance(RamsesInterface, number_of_workers = 3)
        instance.initialize_code()
        error = instance.set_parallel_decomposition(3,1,1)
        self.assertEqual(error, 0)
        instance.setup_mesh(12,20,10,100.0, 300.0, 100.0)
        
        instance.set_boundary("interface","interface","interface","interface","periodic","periodic")
        instance.commit_parameters()
        
           
        for boundaryindex in [3,4]:
            for i0 in range(1,12+4+1):
                for j0 in [1,2]:
                    i = (i0 * 15) + j0
                    error = instance.set_boundary_state(
                        i0,j0,1,       #  index
                        1.0 * (i+1),         #  density
                        2.0 * (i+1), 3.0 * (i+1), 4.0 * (i+1), #  momentum
                        5.0 * (i+1),         #  energy
                        boundaryindex    #  boundary
                    )
                    print(i0, j0)
                    self.assertEqual(error, 0)
                    rho, rhovx, rhovy, rhovz, rhoen, error = instance.get_boundary_state(
                        i0, j0, 1,
                        boundaryindex
                    )
                    self.assertEqual(error, 0)
                    
                    self.assertAlmostRelativeEquals(rho, 1.0 * (i+1))
                    self.assertAlmostRelativeEquals(rhovx, 2.0 * (i+1))
                    self.assertAlmostRelativeEquals(rhovy, 3.0 * (i+1))
                    self.assertAlmostRelativeEquals(rhovz, 4.0 * (i+1))
                    self.assertAlmostRelativeEquals(rhoen, 5.0 * (i+1))
    
    def xtest24(self):
        results = []
        instance=self.new_instance(RamsesInterface, number_of_workers = 3)
        instance.initialize_code()
        error = instance.set_parallel_decomposition(1,3,1)
        self.assertEqual(error, 0)
        instance.setup_mesh(12,30,10,100.0, 300.0, 100.0)
        
        instance.set_boundary("interface","interface","interface","interface","periodic","periodic")
        instance.commit_parameters()
        
           
        for boundaryindex in [3,4]:
            for i0 in range(1,12+4+1):
                for j0 in [1,2]:
                    i = (i0 * 15) + j0
                    error = instance.set_boundary_state(
                        i0,j0,1,       #  index
                        1.0 * (i+1),         #  density
                        2.0 * (i+1), 3.0 * (i+1), 4.0 * (i+1), #  momentum
                        5.0 * (i+1),         #  energy
                        boundaryindex    #  boundary
                    )
                    print(i0, j0)
                    self.assertEqual(error, 0)
                    rho, rhovx, rhovy, rhovz, rhoen, error = instance.get_boundary_state(
                        i0, j0, 1,
                        boundaryindex
                    )
                    self.assertEqual(error, 0)
                    
                    self.assertAlmostRelativeEquals(rho, 1.0 * (i+1))
                    self.assertAlmostRelativeEquals(rhovx, 2.0 * (i+1))
                    self.assertAlmostRelativeEquals(rhovy, 3.0 * (i+1))
                    self.assertAlmostRelativeEquals(rhovz, 4.0 * (i+1))
                    self.assertAlmostRelativeEquals(rhoen, 5.0 * (i+1))
                    
    def xtest25(self):
        results = []
        instance=self.new_instance(RamsesInterface, number_of_workers = 3)
        instance.initialize_code()
        error = instance.set_parallel_decomposition(1,3,1)
        self.assertEqual(error, 0)
        instance.setup_mesh(6,5,5,6.0,5.0,5.0)
        instance.set_boundary("interface","interface","interface","interface","interface","interface")
        instance.commit_parameters()
        
           
        for boundaryindex in [5,6]:
            for i0 in range(1, 6+4+1):
                for j0 in range(1, 5+4+1):
                    for z0 in[1,2]:
                        i = (i0 * (5*4)) + (j0 * 4) + z0
                        error = instance.set_boundary_state(
                            i0,j0,z0,       #  index
                            1.0 * (i+1),         #  density
                            2.0 * (i+1), 3.0 * (i+1), 4.0 * (i+1), #  momentum
                            5.0 * (i+1),         #  energy
                            boundaryindex     #  boundary 
                        )
                        self.assertEqual(error, 0)
                        rho, rhovx, rhovy, rhovz, rhoen, error = instance.get_boundary_state(
                            i0, j0, z0,
                            boundaryindex
                        
                        )
                        self.assertEqual(error, 0)
                        
                        self.assertAlmostRelativeEquals(rho, 1.0 * (i+1))
                        self.assertAlmostRelativeEquals(rhovx, 2.0 * (i+1))
                        self.assertAlmostRelativeEquals(rhovy, 3.0 * (i+1))
                        self.assertAlmostRelativeEquals(rhovz, 4.0 * (i+1))
                        self.assertAlmostRelativeEquals(rhoen, 5.0 * (i+1))
    
    def xtest26(self):
        results = []
        instance=self.new_instance(RamsesInterface, number_of_workers = 9)
        instance.initialize_code()
        error = instance.set_parallel_decomposition(3,3,1)
        instance.setup_mesh(6,6,5,6.0,6.0,5.0)
        self.assertEqual(error, 0)
        instance.set_boundary("interface","interface","interface","interface","interface","interface")
        instance.commit_parameters()
        
           
        for boundaryindex in [5,6]:
            for i0 in range(1,6+4+1):
                for j0 in range(1,6+4+1):
                    for z0 in [1,2]:
                        i = (i0 * (5*4)) + (j0 * 4) + z0
                        error = instance.set_boundary_state(
                            i0,j0,z0,            #  index
                            1.0 * (i+1),         #  density
                            2.0 * (i+1), 3.0 * (i+1), 4.0 * (i+1), #  momentum
                            5.0 * (i+1),         #  energy
                            boundaryindex        #  boundary 
                        )
                        self.assertEqual(error, 0)
                        rho, rhovx, rhovy, rhovz, rhoen, error = instance.get_boundary_state(
                            i0, j0, z0,
                            boundaryindex
                        )
                        self.assertEqual(error, 0)
                        
                        self.assertAlmostRelativeEquals(rho, 1.0 * (i+1))
                        self.assertAlmostRelativeEquals(rhovx, 2.0 * (i+1))
                        self.assertAlmostRelativeEquals(rhovy, 3.0 * (i+1))
                        self.assertAlmostRelativeEquals(rhovz, 4.0 * (i+1))
                        self.assertAlmostRelativeEquals(rhoen, 5.0 * (i+1))
                    
    def xtest27(self):
        instance=RamsesInterface()
        instance.initialize_code()
        gamma, error = instance.get_gamma()
        self.assertAlmostRelativeEquals(gamma, 5.0 / 3.0)
        instance.set_gamma(1.3)
        gamma, error = instance.get_gamma()
        self.assertEqual(error, 0)
        self.assertAlmostRelativeEquals(gamma, 1.3)
        instance.stop()
        
class TestSodShocktube(TestWithMPI):
    
    def xtest0(self):
        N=100
        gamma=5/3.
        g=(gamma-1)/(gamma+1)
        b=(gamma-1)/2/gamma
        
        instance=RamsesInterface()
        instance.initialize_code()
        instance.setup_mesh(N,N/10,N/10,1.,0.1,0.1)
        instance.commit_parameters()
        x,y,z=numpy.indices( (N,N/10,N/10) )
        x=x.flatten()+1
        y=y.flatten()+1
        z=z.flatten()+1
        gamma=5./3.
        rho=0.125*numpy.ones_like(x)
        rhvx=0.*numpy.ones_like(x)
        rhvy=0.*numpy.ones_like(x)
        rhvz=0.*numpy.ones_like(x)
        en=(0.1/(gamma-1))*numpy.ones_like(x)
        instance.set_grid_state(x,y,z,rho,rhvx,rhvy,rhvz,en)
    
        x,y,z=numpy.indices( (N/2,N/10,N/10) )
        x=x.flatten()+1
        y=y.flatten()+1
        z=z.flatten()+1
        rho=1.*numpy.ones_like(x)
        rhvx=0.*numpy.ones_like(x)
        rhvy=0.*numpy.ones_like(x)
        rhvz=0.*numpy.ones_like(x)
        en=(1./(gamma-1))*numpy.ones_like(x)
        instance.set_grid_state(x,y,z,rho,rhvx,rhvy,rhvz,en)
        instance.initialize_grid()
        instance.evolve_model(0.2)
    
        x=numpy.array([0.1,0.9,0.6,0.8])
        y=0.05*numpy.ones_like(x)
        z=0.05*numpy.ones_like(x)
        i,j,k,err=instance.get_index_of_position(x,y,z)
        rho,rhovx,rhovy,rhovz,en,err=instance.get_grid_state(i,j,k)
        vel=numpy.sqrt((rhovx**2+rhovy**2+rhovz**2))/rho
        pres=(gamma-1)*(en-0.5*rho*vel**2)
        u=pres/(gamma-1)/rho
        rhoexp=numpy.zeros_like(x)
        rhoexp[0]=1.
        rhoexp[1]=0.125
        rhoexp[2]=rho[0]*(pres[2]/pres[0])**(1/gamma)
        rhoexp[3]=rho[1]*(pres[3]+g*pres[1])/(pres[1]+g*pres[3])
        
        for i in range(len(rho)):
            self.assertAlmostEqual(rhoexp[i],rho[i],2)
    

class TestRamses(TestWithMPI):
    
    def xtest0(self):
        instance=self.new_instance(Ramses)
        instance.initialize_code()
        instance.stop()
        
    def xtest1(self):
        instance=self.new_instance(Ramses)
        instance.parameters.mesh_size = (10,10,5)
        instance.parameters.length_x = 1.0 | generic_unit_system.length
        instance.parameters.length_y = 1.0 | generic_unit_system.length
        instance.parameters.length_z = 1.0 | generic_unit_system.length
        instance.parameters.x_boundary_conditions = "periodic","periodic"
        instance.parameters.y_boundary_conditions = "periodic","periodic"
        instance.parameters.z_boundary_conditions = "periodic","periodic"
        
    
        self.assertEqual(len(list(instance.itergrids())),1)
        grid = datamodel.Grid(10,10,10)
        grid.rho = 0.4 | generic_unit_system.density
        grid.rhovx = 0.1 | generic_unit_system.momentum_density
        grid.rhovy = 0.0 |  generic_unit_system.momentum_density
        grid.rhovz = 0.0 |  generic_unit_system.momentum_density
        grid.energy = 0.0 | generic_unit_system.energy_density
        
        channel = grid.new_channel_to(instance.grid)
        channel.copy()
            
        instance.initialize_grid()
        
        channel = instance.grid.new_channel_to(grid)
        channel.copy()
        
        self.assertEqual(grid[1][1][0].rho, 0.4 | generic_unit_system.density)
        for x in grid[1].rho.value_in(generic_unit_system.density).flatten():
            self.assertEqual(x, 0.4)
            
        #instance.evolve_model(0.12 | generic_unit_system.time)
        
        #for x in instance.grid.rho.value_in(generic_unit_system.density).flatten():
        #    self.assertEquals(x, 0.1)
    
        #instance.evolve_model(10.0 | generic_unit_system.time)
        #for x in instance.grid.rho.value_in(generic_unit_system.density).flatten():
        #    self.assertEquals(x, 0.1)
        instance.stop()


    def xtest2(self):
        instance=self.new_instance(Ramses)
        instance.parameters.mesh_size = (3,3,3)
        instance.parameters.length_x = 1.0 | generic_unit_system.length
        instance.parameters.length_y = 1.0 | generic_unit_system.length
        instance.parameters.length_z = 1.0 | generic_unit_system.length
        instance.parameters.x_boundary_conditions = "periodic","periodic"
        instance.parameters.y_boundary_conditions = "periodic","periodic"
        instance.parameters.z_boundary_conditions = "periodic","periodic"
        
        grid = datamodel.Grid(3,3,3)
        grid.rho = 0.1 | generic_unit_system.density
        grid.rhovx = 0.0 | generic_unit_system.momentum_density
        grid.rhovy = 0.0 |  generic_unit_system.momentum_density
        grid.rhovz = 0.0 |  generic_unit_system.momentum_density
        grid.energy = 1.0 | generic_unit_system.energy_density
        
        channel = grid.new_channel_to(instance.grid)
        channel.copy()
            
        
        print(instance.grid[1].rho)
        self.assertEqual(instance.grid[1][1][0].rho, 0.1 | generic_unit_system.density)
        for x in instance.grid[1].rho.value_in(generic_unit_system.density).flatten():
            self.assertEqual(x, 0.1)
            
        instance.evolve_model(1.0 | generic_unit_system.time)
        
        for x in instance.grid.rho.value_in(generic_unit_system.density).flatten():
            self.assertEqual(x, 0.1)
    
        instance.evolve_model(10.0 | generic_unit_system.time)
        for x in instance.grid.rho.value_in(generic_unit_system.density).flatten():
            self.assertEqual(x, 0.1)
        instance.stop()
    
    def xtest3(self):
        instance=self.new_instance(Ramses)
        instance.parameters.mesh_size = (5,5,5)
        instance.parameters.length_x = 1.0 | generic_unit_system.length
        instance.parameters.length_y = 1.0 | generic_unit_system.length
        instance.parameters.length_z = 1.0 | generic_unit_system.length
        instance.parameters.x_boundary_conditions = "periodic","periodic"
        instance.parameters.y_boundary_conditions = "periodic","periodic"
        instance.parameters.z_boundary_conditions = "periodic","periodic"
        
    
        grid = datamodel.Grid(5,5,5)
        grid.rho = 0.1 | generic_unit_system.density
        grid.rhovx = 0.0 | generic_unit_system.momentum_density
        grid.rhovy = 0.0 |  generic_unit_system.momentum_density
        grid.rhovz = 0.0 |  generic_unit_system.momentum_density
        grid.energy =  1.0 | generic_unit_system.energy_density
        
        channel = grid.new_channel_to(instance.grid)
        channel.copy()
        
        self.assertEqual((5,5,5), instance.acceleration_grid.shape)
        
        acc_grid = datamodel.Grid(5,5,5)
        acc_grid.ax = 1 | generic_unit_system.acceleration
        acc_grid.ay = 1 | generic_unit_system.acceleration
        acc_grid.az = 1 | generic_unit_system.acceleration
        #self.assertEquals(acc_grid.acceleration[0][0][0], ( 1,1,1) | generic_unit_system.acceleration)
        channel = acc_grid.new_channel_to(instance.acceleration_grid)
        channel.copy()
        
                   
        instance.evolve_model(0.1 | generic_unit_system.time)
        
        self.assertAlmostRelativeEquals(instance.grid.rho, grid.rho);
        self.assertAlmostRelativeEquals(instance.grid.rhovx, 0.1 * 1.0 * 0.1 | generic_unit_system.momentum_density);
        self.assertAlmostRelativeEquals(instance.grid.rhovy, 0.1 * 1.0 * 0.1 | generic_unit_system.momentum_density);
        self.assertAlmostRelativeEquals(instance.grid.rhovz, 0.1 * 1.0 * 0.1 | generic_unit_system.momentum_density);

        instance.evolve_model(0.3 | generic_unit_system.time)
        print(instance.model_time)
        self.assertAlmostRelativeEquals(instance.grid.rho, grid.rho);
        self.assertAlmostRelativeEquals(instance.grid.rhovx, grid.rho *  instance.model_time * acc_grid.ax,2);
        self.assertAlmostRelativeEquals(instance.grid.rhovy, grid.rho *  instance.model_time * acc_grid.ay,2);
        self.assertAlmostRelativeEquals(instance.grid.rhovz, grid.rho *  instance.model_time * acc_grid.az,2);
        instance.stop()
    
    def xtest4(self):
        converter = generic_unit_converter.ConvertBetweenGenericAndSiUnits(
            1 | units.parsec,
            1 | units.Myr,
            1 | units.MSun
        )
        instance=self.new_instance(Ramses, unit_converter = converter)
        instance.parameters.mesh_size = (3,3,3)
        instance.parameters.length_x = 1.0 | units.parsec
        instance.parameters.length_y = 1.0 | units.parsec
        instance.parameters.length_z = 1.0 | units.parsec
        instance.parameters.x_boundary_conditions = "periodic","periodic"
        instance.parameters.y_boundary_conditions = "periodic","periodic"
        instance.parameters.z_boundary_conditions = "periodic","periodic"
        
        instance.commit_parameters()
        density = units.MSun / (units.parsec ** 3)
        grid = datamodel.Grid(3,3,3)
        grid.rho = 0.1 | density
        grid.rhovx = 0.0 | units.MSun / (units.Myr * units.parsec ** 2 )
        grid.rhovy = 0.0 | units.MSun / (units.Myr * units.parsec ** 2 )
        grid.rhovz = 0.0 | units.MSun / (units.Myr * units.parsec ** 2 )
        grid.energy = 1.0 | units.MSun / (units.parsec * units.Myr ** 2)
        
        channel = grid.new_channel_to(instance.grid)
        channel.copy()
            
        
        print(instance.grid[1].rho)
        self.assertAlmostRelativeEquals(instance.grid[1][1][0].rho, 0.1 | density)
        for x in instance.grid[1].rho.value_in(density).flatten():
            self.assertAlmostRelativeEquals(x, 0.1)
            
        instance.evolve_model(1.0 | units.Myr)
        
        for x in instance.grid.rho.value_in(density).flatten():
            self.assertAlmostRelativeEquals(x, 0.1)
    
        instance.evolve_model(10.0 | units.Myr)
        for x in instance.grid.rho.value_in(density).flatten():
            self.assertAlmostRelativeEquals(x, 0.1)
        instance.stop()
    
    
         
    def xtest5(self): 
        instance=self.new_instance(Ramses)
        instance.parameters.mesh_size = (10 , 4, 4)
        instance.parameters.mesh_length = [1.0, 1.0, 1.0] | generic_unit_system.length
        instance.parameters.x_boundary_conditions = ("interface", "outflow")
        instance.parameters.y_boundary_conditions = ("periodic", "periodic")
        instance.parameters.z_boundary_conditions = ("periodic", "periodic")
        instance.parameters.stopping_conditions_number_of_steps = 1
        
        gamma = 5.0 / 3.0
        
        grid = datamodel.new_regular_grid((10,4,4), [1.0, 1.0, 1.0] | generic_unit_system.length )
        
        density = generic_unit_system.density
        momentum =  generic_unit_system.speed * generic_unit_system.density
        energy =  generic_unit_system.mass / ((generic_unit_system.time**2) * generic_unit_system.length)
        
        
        grid.rho = 0.01 | density
        grid.rhovx = 0.1 | momentum
        grid.rhovy = 0.0 | momentum
        grid.rhovz = 0.0 | momentum
        
        p = 1.0 | (generic_unit_system.mass / (generic_unit_system.length * generic_unit_system.time**2))
        
        grid.energy =  p / (gamma - 1)
        grid.energy += 0.5 * (grid.rhovx ** 2  + grid.rhovy ** 2 + grid.rhovz ** 2) / grid.rho
        
        channel = grid.new_channel_to(instance.grid)
        channel.copy()
        instance.stopping_conditions.number_of_steps_detection.enable()
        
        #instance.grid.boundaries.left.
        
        xbound1 = instance.get_boundary_grid('xbound1')
        self.assertEqual(xbound1.shape, (2,4,4))
        memxbound1 = xbound1.copy()
        memxbound1.rho = 0.02 | density
        memxbound1.rhovx = 0.2 | momentum
        memxbound1.rhovy = 0.0 | momentum
        memxbound1.rhovz = 0.0 | momentum
        memxbound1.energy =  p / (gamma - 1)
        memxbound1.energy += 0.5 * (memxbound1.rhovx ** 2  + memxbound1.rhovy ** 2 + memxbound1.rhovz ** 2) / memxbound1.rho
        channel = memxbound1.new_channel_to(xbound1)
        channel.copy()
        
        instance.evolve_model(1.0 | generic_unit_system.time)
        self.assertTrue(instance.stopping_conditions.number_of_steps_detection.is_set())
        
        rho = instance.grid.rho[...,0,0]
        print(rho)
        print(instance.model_time)
        self.assertAlmostRelativeEquals(rho[-1], 0.01 | density)
        self.assertTrue(rho[0] > 0.01 | density)
        self.assertTrue(instance.grid.rhovx[0,0,0] > 0.1 | momentum)
        self.assertAlmostRelativeEquals(instance.grid.rhovx[-1,0,0] , 0.1 | momentum)
        
        instance.stopping_conditions.number_of_steps_detection.disable()
        instance.evolve_model(1.0 | generic_unit_system.time)
        print(instance.model_time)
        rho = instance.grid.rho[...,0,0]
        print(rho)
        self.assertAlmostRelativeEquals(rho, 0.02 | density, 8)
        self.assertAlmostRelativeEquals(instance.grid.rhovx[...,0,0], 0.2 | momentum, 8)
        print(instance.model_time)
        
        instance.stop()
    
    def xtest6(self): 
        instance=self.new_instance(Ramses)
        instance.parameters.mesh_size = (10 , 4, 4)
        instance.parameters.mesh_length = [1.0, 1.0, 1.0] | generic_unit_system.length
        instance.parameters.x_boundary_conditions = ("outflow", "interface")
        instance.parameters.y_boundary_conditions = ("periodic", "periodic")
        instance.parameters.z_boundary_conditions = ("periodic", "periodic")
        instance.parameters.stopping_conditions_number_of_steps = 1
        
        gamma = 5.0 / 3.0
        
        grid = datamodel.new_regular_grid((10,4,4), [1.0, 1.0, 1.0] | generic_unit_system.length )
        
        density = generic_unit_system.density
        momentum =  generic_unit_system.speed * generic_unit_system.density
        energy =  generic_unit_system.mass / ((generic_unit_system.time**2) * generic_unit_system.length)
        
        
        grid.rho = 0.01 | density
        grid.rhovx = -0.1 | momentum
        grid.rhovy = 0.0 | momentum
        grid.rhovz = 0.0 | momentum
        
        p = 1.0 | (generic_unit_system.mass / (generic_unit_system.length * generic_unit_system.time**2))
        
        grid.energy =  p / (gamma - 1)
        grid.energy += 0.5 * (grid.rhovx ** 2  + grid.rhovy ** 2 + grid.rhovz ** 2) / grid.rho
        
        channel = grid.new_channel_to(instance.grid)
        channel.copy()
        instance.stopping_conditions.number_of_steps_detection.enable()
        
        #instance.grid.boundaries.left.
        xbound = instance.get_boundary_grid('xbound2')
        self.assertEqual(xbound.shape, (2,4,4))
        memxbound = xbound.copy()
        memxbound.rho = 0.02 | density
        memxbound.rhovx = -0.2 | momentum
        memxbound.rhovy = 0.0 | momentum
        memxbound.rhovz = 0.0 | momentum
        memxbound.energy =  p / (gamma - 1)
        memxbound.energy += 0.5 * (memxbound.rhovx ** 2  + memxbound.rhovy ** 2 + memxbound.rhovz ** 2) / memxbound.rho
        channel = memxbound.new_channel_to(xbound)
        channel.copy()
        
        instance.evolve_model(1.0 | generic_unit_system.time)
        
        self.assertTrue(instance.stopping_conditions.number_of_steps_detection.is_set())
        rho = instance.grid.rho[...,0,0]
        print(rho)
        print(instance.model_time)
        self.assertAlmostRelativeEquals(rho[0], 0.01 | density)
        self.assertTrue(rho[-1] > 0.01 | density)
        self.assertTrue(instance.grid.rhovx[-1,0,0] < -0.1 | momentum)
        self.assertAlmostRelativeEquals(instance.grid.rhovx[0,0,0] , -0.1 | momentum)
        
        instance.stopping_conditions.number_of_steps_detection.disable()
        instance.evolve_model(1.0 | generic_unit_system.time)
        rho = instance.grid.rho[...,0,0]
        self.assertAlmostRelativeEquals(rho, 0.02 | density, 8)
        self.assertAlmostRelativeEquals(instance.grid.rhovx[...,0,0], -0.2 | momentum, 8)
        print(instance.model_time)
        
        instance.stop()
    
    def xtest7(self): 
        instance=self.new_instance(Ramses, number_of_workers = 2)
        instance.set_parallel_decomposition(1,2,1)
        instance.parameters.mesh_size = (10,4,4)
        instance.parameters.mesh_length = [1.0, 1.0, 1.0] | generic_unit_system.length
        instance.parameters.x_boundary_conditions = ("interface", "outflow")
        instance.parameters.y_boundary_conditions = ("periodic", "periodic")
        instance.parameters.stopping_conditions_number_of_steps = 1
        
        
        gamma = 5.0 / 3.0
        
        grid = datamodel.new_regular_grid((10,4,4), [1.0, 1.0, 1.0] | generic_unit_system.length )
        
        density = generic_unit_system.density
        momentum =  generic_unit_system.speed * generic_unit_system.density
        energy =  generic_unit_system.mass / ((generic_unit_system.time**2) * generic_unit_system.length)
        
        
        grid.rho = 0.01 | density
        grid.rhovx = 0.1 | momentum
        grid.rhovy = 0.0 | momentum
        grid.rhovz = 0.0 | momentum
        
        p = 1.0 | (generic_unit_system.mass / (generic_unit_system.length * generic_unit_system.time**2))
        
        grid.energy =  p / (gamma - 1)
        grid.energy += 0.5 * (grid.rhovx ** 2  + grid.rhovy ** 2 + grid.rhovz ** 2) / grid.rho
        
        channel = grid.new_channel_to(instance.grid)
        channel.copy()
        instance.stopping_conditions.number_of_steps_detection.enable()
        
        #instance.grid.boundaries.left.
        
        xbound1 = instance.get_boundary_grid('xbound1')
        self.assertEqual(xbound1.shape, (2,4,4))
        memxbound1 = xbound1.copy()
        memxbound1.rho = 0.02 | density
        memxbound1.rhovx = 0.2 | momentum
        memxbound1.rhovy = 0.0 | momentum
        memxbound1.rhovz = 0.0 | momentum
        memxbound1.energy =  p / (gamma - 1)
        memxbound1.energy += 0.5 * (memxbound1.rhovx ** 2  + memxbound1.rhovy ** 2 + memxbound1.rhovz ** 2) / memxbound1.rho
        channel = memxbound1.new_channel_to(xbound1)
        channel.copy()
        
        instance.evolve_model(1.0 | generic_unit_system.time)
        self.assertTrue(instance.stopping_conditions.number_of_steps_detection.is_set())
        
        rho = instance.grid.rho[...,0,0]
        print(rho)
        print(instance.model_time)
        self.assertAlmostRelativeEquals(rho[-1], 0.01 | density)
        self.assertTrue(rho[0] > 0.01 | density)
        self.assertTrue(instance.grid.rhovx[0,0,0] > 0.1 | momentum)
        self.assertAlmostRelativeEquals(instance.grid.rhovx[-1,0,0] , 0.1 | momentum)
        
        instance.stopping_conditions.number_of_steps_detection.disable()
        instance.evolve_model(1.0 | generic_unit_system.time)
        print(instance.model_time)
        rho = instance.grid.rho[...,0,0]
        print(rho)
        self.assertAlmostRelativeEquals(rho, 0.02 | density, 8)
        self.assertAlmostRelativeEquals(instance.grid.rhovx[...,0,0], 0.2 | momentum, 8)
        print(instance.model_time)
        
        instance.stop()

    def xtest8(self): 
        instance=self.new_instance(Ramses, number_of_workers = 1)
        #instance.set_parallel_decomposition(2,1,1)
        instance.parameters.mesh_size = (4,10,4)
        instance.parameters.mesh_length = [1.0, 1.0, 1.0] | generic_unit_system.length
        instance.parameters.x_boundary_conditions = ("periodic", "periodic")
        instance.parameters.y_boundary_conditions = ("interface", "outflow")
        instance.parameters.z_boundary_conditions = ("periodic", "periodic")
        instance.parameters.stopping_conditions_number_of_steps = 1
        
        
        gamma = 5.0 / 3.0
        
        grid = datamodel.new_regular_grid((4,10,4), [1.0, 1.0, 1.0] | generic_unit_system.length )
        
        density = generic_unit_system.density
        momentum =  generic_unit_system.speed * generic_unit_system.density
        energy =  generic_unit_system.mass / ((generic_unit_system.time**2) * generic_unit_system.length)
        
        
        grid.rho = 0.01 | density
        grid.rhovx = 0.0 | momentum
        grid.rhovy = 0.1 | momentum
        grid.rhovz = 0.0 | momentum
        
        p = 1.0 | (generic_unit_system.mass / (generic_unit_system.length * generic_unit_system.time**2))
        
        grid.energy =  p / (gamma - 1)
        grid.energy += 0.5 * (grid.rhovx ** 2  + grid.rhovy ** 2 + grid.rhovz ** 2) / grid.rho
        
        channel = grid.new_channel_to(instance.grid)
        channel.copy()
        instance.stopping_conditions.number_of_steps_detection.enable()
        
        ybound = instance.get_boundary_grid('ybound1')
        self.assertEqual(ybound.shape, (4+4,2,4))
        memybound = ybound.copy()
        memybound.rho = 0.02 | density
        memybound.rhovx = 0.0 | momentum
        memybound.rhovy = 0.2 | momentum
        memybound.rhovz = 0.0 | momentum
        memybound.energy =  p / (gamma - 1)
        memybound.energy += 0.5 * (memybound.rhovx ** 2  + memybound.rhovy ** 2 + memybound.rhovz ** 2) / memybound.rho
        
        channel = memybound.new_channel_to(ybound)
        channel.copy()
            
        instance.evolve_model(1.0 | generic_unit_system.time)
        print(instance.stopping_conditions.number_of_steps_detection.is_set())
        
        print(instance.grid.rho[0,...,0])
        rho = instance.grid.rho[0,...,0]
        self.assertAlmostRelativeEquals(rho[-1], 0.01 | density)
        self.assertTrue(rho[0] > 0.01 | density)
        self.assertTrue(instance.grid.rhovy[0,0,0] > 0.1 | momentum)
        self.assertAlmostRelativeEquals(instance.grid.rhovy[0,-1,0] , 0.1 | momentum)
        print(instance.model_time)
        
        instance.stopping_conditions.number_of_steps_detection.disable()
        instance.evolve_model(1.0 | generic_unit_system.time)
        rho = instance.grid.rho[0,...,0]
        self.assertAlmostRelativeEquals(rho, 0.02 | density, 8)
        self.assertAlmostRelativeEquals(instance.grid.rhovy[0,...,0], 0.2 | momentum, 8)
        print(instance.model_time)
        
        instance.stop()

    
    def xtest9(self): 
        instance=self.new_instance(Ramses, number_of_workers = 1)
        #instance.set_parallel_decomposition(2,1,1)
        instance.parameters.mesh_size = (4,10,4)
        instance.parameters.mesh_length = [1.0, 1.0, 1.0] | generic_unit_system.length
        instance.parameters.x_boundary_conditions = ("periodic", "periodic")
        instance.parameters.y_boundary_conditions = ("outflow", "interface")
        instance.parameters.z_boundary_conditions = ("periodic", "periodic")
        instance.parameters.stopping_conditions_number_of_steps = 1
        
        
        gamma = 5.0 / 3.0
        
        grid = datamodel.new_regular_grid((4,10,4), [1.0, 1.0, 1.0] | generic_unit_system.length )
        
        density = generic_unit_system.density
        momentum =  generic_unit_system.speed * generic_unit_system.density
        energy =  generic_unit_system.mass / ((generic_unit_system.time**2) * generic_unit_system.length)
        
        
        grid.rho = 0.01 | density
        grid.rhovx = 0.0 | momentum
        grid.rhovy = -0.1 | momentum
        grid.rhovz = 0.0 | momentum
        
        p = 1.0 | (generic_unit_system.mass / (generic_unit_system.length * generic_unit_system.time**2))
        
        grid.energy =  p / (gamma - 1)
        grid.energy += 0.5 * (grid.rhovx ** 2  + grid.rhovy ** 2 + grid.rhovz ** 2) / grid.rho
        
        channel = grid.new_channel_to(instance.grid)
        channel.copy()
        instance.stopping_conditions.number_of_steps_detection.enable()
        
        ybound = instance.get_boundary_grid('ybound2')
        self.assertEqual(ybound.shape, (4+4,2,4))
        memybound = ybound.copy()
        memybound.rho = 0.02 | density
        memybound.rhovx = 0.0 | momentum
        memybound.rhovy = -0.2 | momentum
        memybound.rhovz = 0.0 | momentum
        memybound.energy =  p / (gamma - 1)
        memybound.energy += 0.5 * (memybound.rhovx ** 2  + memybound.rhovy ** 2 + memybound.rhovz ** 2) / memybound.rho
        
        channel = memybound.new_channel_to(ybound)
        channel.copy()
            
        instance.evolve_model(1.0 | generic_unit_system.time)
        print(instance.stopping_conditions.number_of_steps_detection.is_set())
        print(instance.grid.rho[0,...,0])
        rho = instance.grid.rho[0,...,0]
        self.assertAlmostRelativeEquals(rho[0], 0.01 | density)
        self.assertTrue(rho[-1] > 0.01 | density)
        self.assertTrue(instance.grid.rhovy[0,-1,0] < 0.1 | momentum)
        self.assertAlmostRelativeEquals(instance.grid.rhovy[0,0,0] , -0.1 | momentum)
        print(instance.model_time)
        
        instance.stopping_conditions.number_of_steps_detection.disable()
        instance.evolve_model(1.0 | generic_unit_system.time)
        rho = instance.grid.rho[0,...,0]
        self.assertAlmostRelativeEquals(rho, 0.02 | density, 8)
        self.assertAlmostRelativeEquals(instance.grid.rhovy[0,...,0], -0.2 | momentum, 8)
        print(instance.model_time)
        
        instance.stop()
    
    def xtest10(self): 
        instance=self.new_instance(Ramses, number_of_workers = 1)
        #instance.set_parallel_decomposition(2,1,1)
        instance.parameters.mesh_size = (4,4,10)
        instance.parameters.mesh_length = [1.0, 1.0, 1.0] | generic_unit_system.length
        instance.parameters.x_boundary_conditions = ("periodic", "periodic")
        instance.parameters.y_boundary_conditions = ("periodic", "periodic")
        instance.parameters.z_boundary_conditions = ("interface", "outflow")
        instance.parameters.stopping_conditions_number_of_steps = 1
        
        
        gamma = 5.0 / 3.0
        
        grid = datamodel.new_regular_grid((4,4,10), [1.0, 1.0, 1.0] | generic_unit_system.length )
        
        density = generic_unit_system.density
        momentum =  generic_unit_system.speed * generic_unit_system.density
        energy =  generic_unit_system.mass / ((generic_unit_system.time**2) * generic_unit_system.length)
        
        
        grid.rho = 0.01 | density
        grid.rhovx = 0.0 | momentum
        grid.rhovy = 0.0 | momentum
        grid.rhovz = 0.1 | momentum
        
        p = 1.0 | (generic_unit_system.mass / (generic_unit_system.length * generic_unit_system.time**2))
        
        grid.energy =  p / (gamma - 1)
        grid.energy += 0.5 * (grid.rhovx ** 2  + grid.rhovy ** 2 + grid.rhovz ** 2) / grid.rho
        
        channel = grid.new_channel_to(instance.grid)
        channel.copy()
        instance.stopping_conditions.number_of_steps_detection.enable()
        
        ybound = instance.get_boundary_grid('zbound1')
        self.assertEqual(ybound.shape, (4+4,4+4,2))
        memybound = ybound.copy()
        memybound.rho = 0.02 | density
        memybound.rhovx = 0.0 | momentum
        memybound.rhovy = 0.0 | momentum
        memybound.rhovz = 0.2 | momentum
        memybound.energy =  p / (gamma - 1)
        memybound.energy += 0.5 * (memybound.rhovx ** 2  + memybound.rhovy ** 2 + memybound.rhovz ** 2) / memybound.rho
        
        channel = memybound.new_channel_to(ybound)
        channel.copy()
            
        instance.evolve_model(1.0 | generic_unit_system.time)
        self.assertTrue(instance.stopping_conditions.number_of_steps_detection.is_set())
        
        rho = instance.grid.rho[0,0,...]
        print(rho)
        self.assertAlmostRelativeEquals(rho[-1], 0.01 | density)
        self.assertTrue(rho[0] > 0.01 | density)
        self.assertTrue(instance.grid.rhovz[0,0,0] > 0.1 | momentum)
        self.assertAlmostRelativeEquals(instance.grid.rhovz[0,0,-1] , 0.1 | momentum)
        print(instance.model_time)
        
        instance.stopping_conditions.number_of_steps_detection.disable()
        instance.evolve_model(1.0 | generic_unit_system.time)
        rho = instance.grid.rho[0,0,...]
        self.assertAlmostRelativeEquals(rho, 0.02 | density, 8)
        self.assertAlmostRelativeEquals(instance.grid.rhovz[0,...,0], 0.2 | momentum, 8)
        print(instance.model_time)
        
        instance.stop()

    
    def xtest11(self): 
        instance=self.new_instance(Ramses, number_of_workers = 2)
        instance.set_parallel_decomposition(2,1,1)
        instance.parameters.mesh_size = (4,4,10)
        instance.parameters.mesh_length = [1.0, 1.0, 1.0] | generic_unit_system.length
        instance.parameters.x_boundary_conditions = ("periodic", "periodic")
        instance.parameters.y_boundary_conditions = ("periodic", "periodic")
        instance.parameters.z_boundary_conditions = ("outflow", "interface")
        instance.parameters.stopping_conditions_number_of_steps = 1
        
        
        gamma = 5.0 / 3.0
        
        grid = datamodel.new_regular_grid((4,4,10), [1.0, 1.0, 1.0] | generic_unit_system.length )
        
        density = generic_unit_system.density
        momentum =  generic_unit_system.speed * generic_unit_system.density
        energy =  generic_unit_system.mass / ((generic_unit_system.time**2) * generic_unit_system.length)
        
        
        grid.rho = 0.01 | density
        grid.rhovx = 0.0 | momentum
        grid.rhovy = 0.0 | momentum
        grid.rhovz = -0.10 | momentum
        
        p = 1.0 | (generic_unit_system.mass / (generic_unit_system.length * generic_unit_system.time**2))
        
        grid.energy =  p / (gamma - 1)
        grid.energy += 0.5 * (grid.rhovx ** 2  + grid.rhovy ** 2 + grid.rhovz ** 2) / grid.rho
        
        channel = grid.new_channel_to(instance.grid)
        channel.copy()
        instance.stopping_conditions.number_of_steps_detection.enable()
        
        ybound = instance.get_boundary_grid('zbound2')
        self.assertEqual(ybound.shape, (4+4,4+4,2))
        memybound = ybound.copy()
        memybound.rho = 0.02 | density
        memybound.rhovx = 0.0 | momentum
        memybound.rhovy = 0.0 | momentum
        memybound.rhovz = -0.2 | momentum
        memybound.energy =  p / (gamma - 1)
        memybound.energy += 0.5 * (memybound.rhovx ** 2  + memybound.rhovy ** 2 + memybound.rhovz ** 2) / memybound.rho
        
        channel = memybound.new_channel_to(ybound)
        channel.copy()
            
        instance.evolve_model(1.0 | generic_unit_system.time)
        self.assertTrue(instance.stopping_conditions.number_of_steps_detection.is_set())
        rho = instance.grid.rho[0,0,...]
        self.assertAlmostRelativeEquals(rho[0], 0.01 | density)
        self.assertTrue(rho[-1] > 0.01 | density)
        self.assertTrue(instance.grid.rhovz[0,0,-1] < 0.1 | momentum)
        self.assertAlmostRelativeEquals(instance.grid.rhovz[0,0,0] , -0.1 | momentum)
        print(instance.model_time)
        
        instance.stopping_conditions.number_of_steps_detection.disable()
        instance.evolve_model(1.0 | generic_unit_system.time)
        rho = instance.grid.rho[0,0,...]
        self.assertAlmostRelativeEquals(rho, 0.02 | density, 8)
        self.assertAlmostRelativeEquals(instance.grid.rhovz[0,0,...], -0.2 | momentum, 8)
        print(instance.model_time)
        
        instance.stop()
        
    
    def xtest12(self):
        
        instance=self.new_instance(Ramses)
        instance.parameters.x_boundary_conditions = ("periodic","periodic")
        instance.parameters.y_boundary_conditions = ("periodic","periodic")
        instance.parameters.z_boundary_conditions = ("periodic","periodic")
        instance.parameters.mesh_length = (20.0, 1, 1) | generic_unit_system.length
        instance.parameters.mesh_size = (20, 2, 2)
        
        for x in instance.itergrids():
            inmem = x.copy()
            inmem.rho = inmem.x/(1| generic_unit_system.length) | generic_unit_system.density
            inmem.rhovx = 0.0 | generic_unit_system.momentum_density
            inmem.energy =  1.0 | generic_unit_system.energy_density
            from_model_to_code = inmem.new_channel_to(x)
            from_model_to_code.copy()
            print(inmem.rho)
        rho, rhovx, rhovy, rhovx, rhoenergy = instance.get_hydro_state_at_point(0.5| generic_unit_system.length,0.0| generic_unit_system.length,0.0| generic_unit_system.length)
        
        self.assertEqual(rho , 0.5 | generic_unit_system.density)
        
        for value in numpy.arange(0.5, 19.6, 0.1):
            
            rho, rhovx, rhovy, rhovx, rhoenergy = instance.get_hydro_state_at_point(
                value | generic_unit_system.length,
                0.5 | generic_unit_system.length,
                0.5 | generic_unit_system.length
            )
        
            self.assertAlmostRelativeEquals(rho , value | generic_unit_system.density)
        
        for value in numpy.arange(0.0, 0.6, 0.1):
            rho, rhovx, rhovy, rhovx, rhoenergy = instance.get_hydro_state_at_point(
                value | generic_unit_system.length,
                0.0 | generic_unit_system.length,
                0.0 | generic_unit_system.length
            )
            self.assertAlmostRelativeEquals(rho , ((0.5 + value) * 0.5 + (0.5-value) * 19.5) | generic_unit_system.density)
        
        
        for value in numpy.arange(0.0, 0.5, 0.1):
            rho, rhovx, rhovy, rhovx, rhoenergy = instance.get_hydro_state_at_point(
                value + 19.5| generic_unit_system.length,
                0.0 | generic_unit_system.length,
                0.0 | generic_unit_system.length
            )
            self.assertAlmostRelativeEquals(rho , (19.5 - (value * 19))  | generic_unit_system.density, 9)
        
        # out of range
        rho, rhovx, rhovy, rhovx, rhoenergy = instance.get_hydro_state_at_point(
            20.0| generic_unit_system.length,
            0.0 | generic_unit_system.length,
            0.0 | generic_unit_system.length
        )
        self.assertAlmostRelativeEquals(rho , 0.0 | generic_unit_system.density, 9)

    def xtest13(self):
        
        instance=self.new_instance(Ramses,  number_of_workers=2)
        instance.parameters.x_boundary_conditions = ("periodic","periodic")
        instance.parameters.y_boundary_conditions = ("periodic","periodic")
        instance.parameters.y_boundary_conditions = ("periodic","periodic")
        instance.parameters.mesh_length = (20.0, 20.0, 4) | generic_unit_system.length
        instance.parameters.mesh_length = (20.0, 20.0, 4) | generic_unit_system.length
        instance.parameters.mesh_size = (20, 20, 2)
        
        for x in instance.itergrids():
            inmem = x.copy()
            inmem.rho = (inmem.x + ((inmem.y - (0.5| generic_unit_system.length))* 20.0))/(1| generic_unit_system.length) | generic_unit_system.density
            inmem.rhovx = 0.0 | generic_unit_system.momentum_density
            inmem.energy =  1.0 | generic_unit_system.energy_density
            from_model_to_code = inmem.new_channel_to(x)
            from_model_to_code.copy()
            print(inmem.rho[0], inmem.y[0], inmem.x[0])
        rho, rhovx, rhovy, rhovx, rhoenergy = instance.get_hydro_state_at_point(0.5| generic_unit_system.length,0.5| generic_unit_system.length,0.0| generic_unit_system.length)
        
        self.assertEqual(rho , 0.5 | generic_unit_system.density)
        
        for value in numpy.arange(0.5, 19.6, 0.1):
            
            rho, rhovx, rhovy, rhovx, rhoenergy = instance.get_hydro_state_at_point(
                value | generic_unit_system.length,
                0.5 | generic_unit_system.length,
                0.0 | generic_unit_system.length
            )
        
            self.assertAlmostRelativeEquals(rho , value | generic_unit_system.density)
        
        for x in numpy.arange(8.5, 11.5, 0.25):
            for y in numpy.arange(0.5, 19.6, 0.25):
                rho, rhovx, rhovy, rhovx, rhoenergy = instance.get_hydro_state_at_point(
                    x | generic_unit_system.length,
                    y | generic_unit_system.length,
                    0.0 | generic_unit_system.length
                )
            
                self.assertAlmostRelativeEquals(rho , x + (20 * (y-0.5))  | generic_unit_system.density)
            
    def xtest14(self):
        
        instance=self.new_instance(Ramses, number_of_workers=3)
        instance.parameters.x_boundary_conditions = ("periodic","periodic")
        instance.parameters.y_boundary_conditions = ("periodic","periodic")
        instance.parameters.z_boundary_conditions = ("periodic","periodic")
        instance.parameters.mesh_length = (20.0, 20.0, 20.0) | generic_unit_system.length
        instance.parameters.mesh_length = (20.0, 20.0, 20.0) | generic_unit_system.length
        instance.parameters.mesh_size = (20, 20, 20)
        
        for x in instance.itergrids():
            inmem = x.copy()
            inmem.rho = (
                (
                    inmem.x + 
                    ((inmem.y - (0.5| generic_unit_system.length))* 20.0) +
                    ((inmem.z - (0.5| generic_unit_system.length))* 400.0)
                )
                /(1| generic_unit_system.length) | generic_unit_system.density
            )
            inmem.rhovx = 0.0 | generic_unit_system.momentum_density
            inmem.energy =  1.0 | generic_unit_system.energy_density
            from_model_to_code = inmem.new_channel_to(x)
            from_model_to_code.copy()
        rho, rhovx, rhovy, rhovx, rhoenergy = instance.get_hydro_state_at_point(0.5| generic_unit_system.length,0.5| generic_unit_system.length,0.5| generic_unit_system.length)
        
        self.assertEqual(rho , 0.5 | generic_unit_system.density)
        
        for value in numpy.arange(0.5, 19.6, 0.1):
            
            rho, rhovx, rhovy, rhovx, rhoenergy = instance.get_hydro_state_at_point(
                value | generic_unit_system.length,
                0.5 | generic_unit_system.length,
                0.5 | generic_unit_system.length
            )
        
            self.assertAlmostRelativeEquals(rho , value | generic_unit_system.density)
        
        sample = datamodel.new_regular_grid(
            (4, 4, 76),
            (2, 2, 19) | generic_unit_system.length
        )
        sample.x += 9.5 | generic_unit_system.length
        sample.y += 9.5 | generic_unit_system.length
        sample.z += 0.5 | generic_unit_system.length
        x = sample.x.flatten()
        y = sample.y.flatten()
        z = sample.z.flatten()
        
        rho, rhovx, rhovy, rhovx, rhoenergy = instance.get_hydro_state_at_point(
            x,
            y,
            z
        )
        half = 0.5 | generic_unit_system.length
        
        self.assertAlmostRelativeEquals(rho , (x + (20 * (y-half)) + (400 * (z-half)))/(1| generic_unit_system.length) | generic_unit_system.density )
            

    
    def xtest15(self): 
        instance=self.new_instance(Ramses, number_of_workers = 1)
        self.assertAlmostRelativeEquals(instance.parameters.gamma, 5.0 / 3.0)
        instance.parameters.gamma = 1.2
        self.assertAlmostRelativeEquals(instance.parameters.gamma, 1.2)
        #self.assertAlmostRelativeEquals(instance.parameters.timestep, 0.1 | generic_unit_system.time)
        #instance.parameters.timestep = 0.2 | generic_unit_system.time
        #self.assertAlmostRelativeEquals(instance.parameters.timestep, 0.2 | generic_unit_system.time)



from amuse.support.interface import InCodeComponentImplementation

from amuse.test.amusetest import TestWithMPI

from amuse import datamodel
from amuse.rfi.core import PythonCodeInterface, legacy_function,LegacyFunctionSpecification

from amuse.support import exceptions

import numpy
from mpi4py import MPI

class ForTestingInterface(PythonCodeInterface):
    
    def __init__(self, **options):
        PythonCodeInterface.__init__(self, implementation_factory = ForTestingImplementation, **options)
        
    @legacy_function
    def get_range():
        function = LegacyFunctionSpecification()  
        function.addParameter('imin', dtype='int32', direction=function.OUT)
        function.addParameter('imax', dtype='int32', direction=function.OUT)
        function.result_type = 'int32'
        return function
        
    @legacy_function
    def get_x():
        function = LegacyFunctionSpecification()  
        function.addParameter('index', dtype='int32', direction=function.IN)
        function.addParameter('x', dtype='float64', direction=function.OUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function    

    @legacy_function
    def get_dens():
        function = LegacyFunctionSpecification()  
        function.addParameter('index', dtype='int32', direction=function.IN)
        function.addParameter('dens', dtype='float64', direction=function.OUT)
        function.addParameter('N', dtype='int32', direction=function.LENGTH)

        function.result_type = 'int32'
        function.must_handle_array = True
        return function    

class ForTestingImplementation(object):
    
    def __init__(self):
        self.comm=MPI.COMM_WORLD
        self.myrank=self.comm.Get_rank()
        self.N=self.comm.Get_size()
        self.Ngrid=3*4*5
        n=self.Ngrid//self.N
        x = (numpy.arange(n)+self.myrank*n)/(1.*self.Ngrid)
        self.local_imin=self.myrank*n
        self.local_imax=(self.myrank+1)*n-1
        self.dens = x**2
        
    def get_range(self,imin,imax):
        imin.value=0
        imax.value=self.Ngrid-1
        return 0
        
    def get_x(self,index,x):
        x.value=index/(1.*self.Ngrid)
        return 0

    def get_dens(self,index,dens,N):
        a=(index>=self.local_imin)*(index<=self.local_imax)
        _dens=numpy.zeros(N)
        _dens[a]=self.dens[index[a]-self.local_imin]
        dens.value=numpy.zeros(N)
        _dens=self.comm.Reduce(_dens, dens.value, MPI.SUM,root=0)
        return 0

class ForTesting(InCodeComponentImplementation):
    
    def __init__(self, **options):
        InCodeComponentImplementation.__init__(self, ForTestingInterface(**options), **options)

    def define_grids(self, object):        
        object.define_grid('grid',axes_names = ['x'],grid_class=datamodel.CartesianGrid)
        object.set_grid_range('grid', 'get_range')
        object.add_getter('grid', 'get_dens', names=('dens',))
        object.add_getter('grid', 'get_x', names=('x',))
            
class TestInterface(TestWithMPI):
    
    def ForTesting(self, **options):
        options["worker_dir"]=self.get_path_to_results()
        return ForTesting( **options)
    def ForTestingInterface(self, **options):
        options["worker_dir"]=self.get_path_to_results()
        return ForTestingInterface(**options)

    def test1(self):
        interface=self.ForTesting(redirection="none",number_of_workers=1)        
        x=interface.grid.x
        dens=interface.grid.dens
        self.assertEqual(x,numpy.arange(60)/60.)
        self.assertEqual(dens,x**2)
        interface.stop()
    
    def test2(self):
        for n in [3,5,6]:
            interface=self.ForTesting(redirection="none",number_of_workers=n)        
            x=interface.grid.x
            dens=interface.grid.dens
            self.assertEqual(x,numpy.arange(60)/60.)
            self.assertEqual(dens,x**2)
            interface.stop()




import os
import os.path
import numpy

from amuse.units import units
from amuse.datamodel import Particles,Particle
from amuse.support.exceptions import AmuseException
from amuse.test.amusetest import TestCase, get_path_to_results

from amuse.community.mesa.interface import MESA
from amuse.community.evtwin.interface import EVtwin
from amuse.community.sse.interface import SSE

from amuse.couple.fallback_stellar_evolution import FallbackStellarEvolution

class TestFallbackStellarEvolution(TestCase):

    def test1(self):
        print("Testing FallbackStellarEvolution")
        instance = FallbackStellarEvolution()
        instance.stop()

    def xtest2(self):
        print("Testing FallbackStellarEvolution: evolve tests")


# results of original code  (not really check the numbers, tests have been relaxed because
# different timestepping of evolve_model and evolve_one_step)
        results=dict()
        results["10.0 MSun"]=dict(sse_age=3644655.52487 | units.yr,
                                      sse_mass=9.99289747724 | units.MSun,
                                      sse_rad=4.24764389443 | units.RSun,
                                      sse_L=5993.72678228 | units.LSun,
                                      evtwin_age=2999999.99732 | units.yr,
                                      evtwin_mass=9.99832250931 | units.MSun,
                                      evtwin_rad=4.1727904434 | units.RSun,
                                      evtwin_L=6175.5214329 | units.LSun)
        results["20.0 MSun"]=dict(sse_age= 2609167.84395| units.yr,
                                      sse_mass= 19.8895071502| units.MSun,
                                      sse_rad=7.07713598137| units.RSun,
                                      sse_L=51232.436257| units.LSun,
                                      evtwin_age= 2999999.99792| units.yr,
                                      evtwin_mass=19.877338008| units.MSun,
                                      evtwin_rad=6.77369109826| units.RSun,
                                      evtwin_L= 54123.7589452| units.LSun)
        results["40.0 MSun"]=dict(sse_age= 2627093.59096| units.yr,
                                      sse_mass=38.3702723816| units.MSun,
                                      sse_rad=13.6137662257 | units.RSun,
                                      sse_L=303498.877464 | units.LSun,
                                      evtwin_age= 2999999.99923| units.yr,
                                      evtwin_mass=37.0862227121 | units.MSun,
                                      evtwin_rad= 13.5982316053| units.RSun,
                                      evtwin_L=320099.608846| units.LSun)


        for m in [40.]| units.MSun:#,20.,10.] | units.MSun:
          instance = FallbackStellarEvolution()
          instance._main_se.parameters.max_age_stop_condition=3.| units.Myr

          star = Particle(1)
          star.mass = m
          star = instance.particles.add_particle(star)
          instance.commit_particles()

          while instance.ActiveModel[star].__class__.__name__=="EVtwin":
              instance.evolve_model()

          print("%s\t%s\t%s\t%s\t%s\t%s" % (star.age, star.mass, star.radius,
           star.luminosity, star.stellar_type, instance.ActiveModel[star].__class__.__name__))

          self.assertAlmostRelativeEqual(results[str(m)]["sse_age"],instance._FBTimeseries[star].particles[0].SSEAgeAtSwitch,7)
          self.assertAlmostRelativeEqual(results[str(m)]["sse_mass"],star.mass,7)
          self.assertAlmostRelativeEqual(results[str(m)]["sse_rad"],star.radius,7)
          self.assertAlmostRelativeEqual(results[str(m)]["sse_L"],star.luminosity,7)


          star=instance._main_se.particles[0]

          self.assertAlmostRelativeEqual(results[str(m)]["evtwin_age"],star.age,7)
          self.assertAlmostRelativeEqual(results[str(m)]["evtwin_mass"],star.mass,2)
          self.assertAlmostRelativeEqual(results[str(m)]["evtwin_rad"],star.radius,2)
          self.assertAlmostRelativeEqual(results[str(m)]["evtwin_L"],star.luminosity,2)

          print("%s\t%s\t%s\t%s\t%s" % (star.age, star.mass, star.radius,
           star.luminosity, star.stellar_type))

          instance.stop()

    def slowtest3(self):
        print("Testing FallbackStellarEvolution: evolve 3 stars at the same time")


# results of original code  (not really check the numbers, tests have been relaxed because
# different timestepping of evolve_model and evolve_one_step)
        results=dict()
        results["10.0 MSun"]=dict(sse_age=3644655.52487 | units.yr,
                                      sse_mass=9.99289747724 | units.MSun,
                                      sse_rad=4.35949272485 | units.RSun,
                                      sse_L=6180.22077675| units.LSun,
                                      evtwin_age=2999999.99732 | units.yr,
                                      evtwin_mass=9.99832250931 | units.MSun,
                                      evtwin_rad=4.1727904434 | units.RSun,
                                      evtwin_L=6175.5214329 | units.LSun)
        results["20.0 MSun"]=dict(sse_age= 2609167.84395| units.yr,
                                      sse_mass= 19.8895071502| units.MSun,
                                      sse_rad=7.30432858615| units.RSun,
                                      sse_L=51232.436257| units.LSun,
                                      evtwin_age= 2999999.99792| units.yr,
                                      evtwin_mass=19.877338008| units.MSun,
                                      evtwin_rad=6.77369109826| units.RSun,
                                      evtwin_L= 54123.7589452| units.LSun)
        results["40.0 MSun"]=dict(sse_age= 2869735.64001| units.yr, # for some reason this is not 2617093.59096
                                      sse_mass=37.9172423183| units.MSun,
                                      sse_rad=15.1388511451| units.RSun,
                                      sse_L=322674.980598| units.LSun,
                                      evtwin_age= 2999999.99923| units.yr,
                                      evtwin_mass=37.0862227121 | units.MSun,
                                      evtwin_rad= 13.5982316053| units.RSun,
                                      evtwin_L=320099.608846| units.LSun)


        instance = FallbackStellarEvolution()
        instance._main_se.parameters.max_age_stop_condition=3.| units.Myr

        stars = Particles(3,mass=[10.,20.,40.] | units.MSun)
        stars = instance.particles.add_particles(stars)
        stars.initial_mass=stars.mass
        instance.commit_particles()

        while instance.model_time<.05| units.Myr:
              instance.evolve_model()

        for star in stars:
          self.assertEqual(instance.ActiveModel[star].__class__.__name__,"EVtwin")

        while instance.model_time<=3.| units.Myr:
              instance.evolve_model()

        for star in stars:
          self.assertEqual(instance.ActiveModel[star].__class__.__name__,"SSE")

        for star in stars:
          m=star.initial_mass
          self.assertAlmostRelativeEqual(results[str(m)]["sse_age"],instance._FBTimeseries[star].particles[0].SSEAgeAtSwitch,3)

          evstar=star.as_particle_in_set(instance._main_se.particles)

          self.assertAlmostRelativeEqual(results[str(m)]["evtwin_age"],evstar.age,7)
          self.assertAlmostRelativeEqual(results[str(m)]["evtwin_mass"],evstar.mass,2)
          self.assertAlmostRelativeEqual(results[str(m)]["evtwin_rad"],evstar.radius,2)
          self.assertAlmostRelativeEqual(results[str(m)]["evtwin_L"],evstar.luminosity,2)


        instance.stop()

    def slowtest4(self):
        print("Testing FallbackStellarEvolution: evolve same 3 stars at the same time")

        instance = FallbackStellarEvolution()
        instance._main_se.parameters.max_age_stop_condition=3.| units.Myr

        stars = Particles(3,mass=[40.,40.,40.] | units.MSun)
        stars = instance.particles.add_particles(stars)
        instance.commit_particles()

        while instance.model_time<=3.| units.Myr:
              instance.evolve_model()

        for star in stars:
          self.assertEqual(instance.ActiveModel[star].__class__.__name__,"SSE")

        age1=instance._FBTimeseries[stars[0]].particles[0].SSEAgeAtSwitch
        age2=instance._FBTimeseries[stars[1]].particles[0].SSEAgeAtSwitch
        age3=instance._FBTimeseries[stars[2]].particles[0].SSEAgeAtSwitch

        self.assertEqual(age1,age2)
        self.assertEqual(age2,age3)

        instance.stop()

    def slowtest5(self):
        print("Testing FallbackStellarEvolution: evolve with end time")

        instance = FallbackStellarEvolution()
        instance._main_se.parameters.max_age_stop_condition=0.1| units.Myr

        stars = Particles(3,mass=[10.,20.,30.] | units.MSun)
        stars = instance.particles.add_particles(stars)
        instance.commit_particles()

        instance.evolve_model(.101 | units.Myr)

        for star in stars:
          self.assertEqual(instance.ActiveModel[star].__class__.__name__,"SSE")

        self.assertTrue(.101| units.Myr<instance.model_time)

        instance.stop()

    def test6(self):
        print("Testing FallbackStellarEvolution: enforce monotonic mass evolution")

        instance = FallbackStellarEvolution(enforce_monotonic_mass_evolution=True)
        instance._main_se.parameters.max_age_stop_condition=3.| units.Myr

        stars = Particles(1,mass=[40.] | units.MSun)
        stars = instance.particles.add_particles(stars)
        instance.commit_particles()

        while instance.ActiveModel[stars[0]].__class__.__name__=="EVtwin":
          emass=stars[0].mass
          instance.evolve_model()

        self.assertTrue(stars[0].mass<=emass)

        instance.stop()

    def slowtest7(self):
        print("Testing FallbackStellarEvolution: test mesa")

        instance = FallbackStellarEvolution(MESA)
        instance._main_se.parameters.max_age_stop_condition=3.| units.Myr

        stars = Particles(3,mass=[40.,40.,40.] | units.MSun)
        stars = instance.particles.add_particles(stars)
        instance.commit_particles()

        instance.evolve_model(3.01 | units.Myr)

        for star in stars:
          self.assertEqual(instance.ActiveModel[star].__class__.__name__,"SSE")

        age1=instance._FBTimeseries[stars[0]].particles[0].SSEAgeAtSwitch
        age2=instance._FBTimeseries[stars[1]].particles[0].SSEAgeAtSwitch
        age3=instance._FBTimeseries[stars[2]].particles[0].SSEAgeAtSwitch

        self.assertEqual(age1,age2)
        self.assertEqual(age2,age3)

        instance.stop()





from amuse.test import amusetest

from amuse.datamodel.grids import *

import numpy

class TestGridAttributes(amusetest.TestCase):
    
    def test1(self):
        grid=new_cartesian_grid((10,20),1.)
        self.assertEqual(grid.cellsize(),[1.,1.])
        self.assertEqual(grid.get_minimum_index(), (0,0))
        self.assertEqual(grid.get_maximum_index(), (9,19))
        self.assertEqual(grid.get_minimum_position(), [0.,0.])
        self.assertEqual(grid.get_maximum_position(), [10.,20.])
        self.assertEqual(grid.get_volume(), 10*20.)
        self.assertEqual(grid.contains(numpy.asarray([[1,1],[-10,-10]])), [True,False])
        points=grid.points()
        x=points[:,:,0]
        y=points[:,:,1]
        self.assertEqual([x,y],numpy.indices((11,21)))

        grid2=new_cartesian_grid((10,20),1., offset=(5,5))
        self.assertTrue(grid.overlaps(grid2))
        grid3=new_cartesian_grid((10,20),1., offset=(15,25))
        self.assertFalse(grid.overlaps(grid3))
        overlap=grid.get_overlap_with(grid2)
        self.assertEqual(overlap.shape,(5,15))
    
    def test2(self):
        grid=new_regular_grid((10,20),(10,10))
        self.assertEqual(grid.cellsize(),[1.,0.5])
        self.assertEqual(grid.get_minimum_index(), (0,0))
        self.assertEqual(grid.get_maximum_index(), (9,19))
        self.assertEqual(grid.get_minimum_position(), [0.,0.])
        self.assertEqual(grid.get_maximum_position(), [10.,10.])
        self.assertEqual(grid.get_volume(), 10*10.)
        self.assertEqual(grid.contains(numpy.asarray([[1,1],[-10,-10]])), [True,False])
        points=grid.points()
        x=points[:,:,0]
        y=points[:,:,1]
        xp,yp=numpy.indices((11,21))
        self.assertEqual(x,xp)
        self.assertEqual(y,yp*0.5)

        grid2=new_cartesian_grid((10,20),1., offset=(5,5))
        self.assertTrue(grid.overlaps(grid2))
        grid3=new_cartesian_grid((10,20),1., offset=(15,25))
        self.assertFalse(grid.overlaps(grid3))
        overlap=grid.get_overlap_with(grid2)
        self.assertEqual(overlap.shape,(5,10))


    def test3(self):
        x=numpy.arange(11)/10.
        y=numpy.arange(21)/20.
        grid=new_rectilinear_grid((10,20),(x,y**2))
        self.assertRaises(Exception, grid.cellsize,
          expected_message="a RectilinearGrid does not have a constant cellsize, use the cellsizes method instead")

    def xtest4(self):
        grid=new_structured_grid((10,20),(10,10))
        self.assertEqual(grid.cellsize(),[1.,0.5])

    def xtest5(self):
        grid=new_unstructured_grid((10,20),(10,10))
        self.assertEqual(grid.cellsize(),[1.,0.5])

    def test6(self):
        grid=new_regular_grid((10,10),(10,10))
        self.assertEqual(grid.get_index((6.2,3.7)), [6,3])
        self.assertEqual(grid.get_index(x=[6.2],y=[3.7]), [6,3])
        self.assertEqual(grid.get_index(y=[6.2],x=[3.7]), [3,6])
        self.assertEqual(grid.get_index(y=6.2,x=3.7)[0], 3)
        self.assertEqual(grid.get_index(y=6.2,x=3.7)[1], 6)
        
    def test7(self):
        grid=new_regular_grid((10,10),[20,10] | units.m,axes_names="ab")
        self.assertEqual(grid.get_index([16.2,3.7] | units.m), [8,3])
        self.assertEqual(grid.get_index(a=16.2 | units.m,b=3.7 | units.m), [8,3])
        self.assertEqual(grid.get_index(a=[16.2, 4.5] | units.m,b=[3.7,4.2] | units.m), [[8,3],[2,4]])




from amuse.datamodel import AbstractSet
from amuse.test.amusetest import TestWithMPI
from amuse.ext import concurrent
from amuse import datamodel

from amuse.rfi.core import *
from amuse.ic.plummer import new_plummer_sphere
from amuse.units import nbody_system
from amuse.units import units

from amuse.community import *
from amuse.support.interface import InCodeComponentImplementation

from amuse.io import read_set_from_file, write_set_to_file

from collections import namedtuple
import numpy
import pickle
try:
    from mpi4py import MPI
except ImportError:
    MPI = None
    
import base64

def dump_and_encode(x):
  return base64.b64encode(pickle.dumps(x)).decode()
def decode_and_load(x):
  return pickle.loads(base64.b64decode(x.encode()))

class DistributedParticlesInterface(PythonCodeInterface):
    
    def __init__(self, implementation_factory, **options):
        PythonCodeInterface.__init__(self, implementation_factory = implementation_factory, **options)
        
    @legacy_function
    def get_length():
        function = LegacyFunctionSpecification()  
        function.addParameter('reference', dtype='int64', direction=function.IN)
        function.addParameter('len_out', dtype='int64', direction=function.OUT)
        function.result_type = 'int32'
        return function
        
    @legacy_function
    def do_init():
        function = LegacyFunctionSpecification()  
        function.addParameter('size', dtype='int64', direction=function.IN)
        function.addParameter('reference_out', dtype='int64', direction=function.OUT)
        function.result_type = 'int32'
        return function
        
    @legacy_function
    def do_set_attribute():
        function = LegacyFunctionSpecification()  
        function.addParameter('reference', dtype='int64', direction=function.IN)
        function.addParameter('name_of_the_attribute', dtype='string', direction=function.IN)
        function.addParameter('pickled_value', dtype='string', direction=function.IN)
        function.result_type = 'int32'
        return function
        
    @legacy_function
    def do_get_attribute():
        function = LegacyFunctionSpecification()  
        function.addParameter('reference', dtype='int64', direction=function.IN)
        function.addParameter('name_of_the_attribute', dtype='string', direction=function.IN)
        function.addParameter('output', dtype='string', direction=function.OUT)
        function.result_type = 'int32'
        return function
        
    @legacy_function
    def do_getitem():
        function = LegacyFunctionSpecification()  
        function.addParameter('reference_in', dtype='int64', direction=function.IN)
        function.addParameter('pickled_index', dtype='string', direction=function.IN)
        function.addParameter('is_particle_out', dtype='bool', direction=function.OUT)
        function.addParameter('reference_out', dtype='int64', direction=function.OUT)
        function.result_type = 'int32'
        return function

    @legacy_function
    def set_from_generator():
        function = LegacyFunctionSpecification()  
        function.addParameter('reference_in', dtype='int64', direction=function.IN)
        function.addParameter('pickled_generator', dtype='string', direction=function.IN)
        function.addParameter('pickled_args', dtype='string', direction=function.IN)
        function.addParameter('pickled_kwargs', dtype='string', direction=function.IN)
        function.result_type = 'int32'
        return function        

    @legacy_function
    def select_array():
        function = LegacyFunctionSpecification()  
        function.addParameter('reference_in', dtype='int64', direction=function.IN)
        function.addParameter('pickled_function', dtype='string', direction=function.IN)
        function.addParameter('pickled_attributes', dtype='string', direction=function.IN)
        function.addParameter('reference_out', dtype='int64', direction=function.OUT)
        function.result_type = 'int32'
        return function

    @legacy_function
    def write_set_to_file():
        function = LegacyFunctionSpecification()  
        function.addParameter('reference_in', dtype='int64', direction=function.IN)
        function.addParameter('filebase', dtype='string', direction=function.IN)
        function.addParameter('fileformat', dtype='string', direction=function.IN)
        function.addParameter('pickled_filenames', dtype='string', direction=function.OUT)
        function.result_type = 'int32'
        return function

    @legacy_function
    def read_set_from_file():
        function = LegacyFunctionSpecification()  
        function.addParameter('reference_in', dtype='int64', direction=function.IN)
        function.addParameter('pickled_filenames', dtype='string', direction=function.IN)
        function.addParameter('fileformat', dtype='string', direction=function.IN)
        function.result_type = 'int32'
        return function        
        
  
class DistributedParticlesCode(InCodeComponentImplementation):
    
    def __init__(self, **options):
        interface = DistributedParticlesInterface(
            implementation_factory = DistributedParticlesImplementation,
            **options
        )
        
        InCodeComponentImplementation.__init__(
            self,
            interface,
            **options
        )
        
class DistributedParticles(object):
    
    def __init__(self, size = 0, reference = 0, code = None, **options):
        if code is None:
            code = DistributedParticlesCode(**options)
            code.do_init(size)
            reference = 0
        
        object.__setattr__(self, "code", code)
        object.__setattr__(self, "reference", reference)
        
    def __len__(self):
        return self.code.get_length(self.reference)
        
    def __setattr__(self, name_of_the_attribute, value):
        self.code.do_set_attribute(
            self.reference,
            name_of_the_attribute, 
            dump_and_encode(value)
        )
        
    def __getattr__(self, name_of_the_attribute):
        result=self.code.do_get_attribute(self.reference,name_of_the_attribute)
        return decode_and_load(result)
        
    def __getitem__(self, index):
        is_particle, reference = self.code.do_getitem(
            self.reference,
            dump_and_encode(index)
        )
        if is_particle:
            return DistributedParticle(
                code = self.code,
                reference = reference
            )
        else:
            return DistributedParticles(
                code = self.code,
                reference = reference
            )

    def set_from_generator(self,func, args=(),kwargs={}):        
        self.code.set_from_generator(self.reference, dump_and_encode(func),
                dump_and_encode(args),dump_and_encode(kwargs))

    def select_array(self, function, attributes):
        
        reference=self.code.select_array(self.reference,
                                         dump_and_encode(function),
                                         dump_and_encode(attributes)
        )

        return DistributedParticles(
            code = self.code,
            reference = reference
        )

    def write_set_to_file(self, filebase, fileformat):
      
        filenames=self.code.write_set_to_file(self.reference,filebase,fileformat)
        
        return decode_and_load(filenames)

    def read_set_from_file(self, filenames, fileformat):
        if self.reference!=0:
            raise Exception("read only allowed to root set") 
        
        self.code.read_set_from_file(self.reference,
                                     dump_and_encode(filenames), 
                                     fileformat )
            
class DistributedParticle(object):
    
    def __init__(self, reference, code, **options):
        object.__setattr__(self, "code", code)
        object.__setattr__(self, "reference", reference)
        
    def __setattr__(self, name_of_the_attribute, value):
        self.code.do_set_attribute(
            self.reference,
            name_of_the_attribute, 
            dump_and_encode(value)
        )
        
    def __getattr__(self, name_of_the_attribute):
        return decode_and_load(self.code.do_get_attribute(self.reference,name_of_the_attribute))
        
ReferencedParticles = namedtuple('ReferencedParticles', ['particles', 'local_offset', 'local_size', 'global_size'])

class DistributedParticlesImplementation(object):
    EmptyReference = ReferencedParticles(
        None,
        0,
        0,
        0
    )
    
    def do_init(self, size, reference_out):
        
        self.mpi_comm = MPI.COMM_WORLD
        self.rank =  MPI.COMM_WORLD.Get_rank()
        self.number_of_processes = MPI.COMM_WORLD.Get_size()
        local_size  = size // self.number_of_processes 
        left_over = size - (local_size * self.number_of_processes)
        if self.rank == 0:
            local_size += left_over
            local_offset = 0
        else:
            local_offset = (local_size * self.rank) + left_over
        self.reference_counter = 0
        self.references_to_particles = {}
        real_particles = datamodel.Particles(local_size)
        self.references_to_particles[self.reference_counter] = ReferencedParticles(
            real_particles,
            local_offset,
            local_size,
            size
        )
        reference_out.value = self.reference_counter
        self.reference_counter += 1
        return 0
    
    def get_length(self, reference, len_out):
        particles_len = self.references_to_particles[reference].local_size
        
        input = numpy.zeros(1,  dtype='int64')
        output = numpy.zeros(1,  dtype='int64')
        
        input[0] = particles_len
            
        self.mpi_comm.Reduce(
            [input, MPI.INTEGER], 
            [output, MPI.INTEGER],
            op=MPI.SUM, 
            root=0
        )
        len_out.value = output[0]
        return 0
        
    def do_set_attribute(self, reference, name_of_the_attribute, pickled_value):
        real_particles,local_offset, local_size, global_size = self.references_to_particles[reference]
        value = decode_and_load(pickled_value)
        if not real_particles is None:
            if (is_quantity(value) and not value.is_scalar()) or hasattr(value, '__iter__'):
                setattr(real_particles, name_of_the_attribute, value[local_offset:local_offset + local_size])
            else:
                setattr(real_particles, name_of_the_attribute, value)
        return 0
        
    def _merge_results(self, quantities): 
        values = []
        value = None
        quantities = [x for x in quantities if x[-1] is not None]
        
        if len(quantities) == 0:
            return None
            
        if len(quantities) == 1:
            local_size, local_offset, quantity = quantities[0]
            if local_size == 0:
                return quantity
        
        total_size = 0
        for local_size, local_offset, quantity in quantities:
            total_size += local_size
            
        first_quantity = quantities[0][-1]
        result = numpy.zeros(total_size, dtype = first_quantity.dtype)
        if is_quantity(first_quantity):
            result = result | first_quantity.unit
            
        for local_size, local_offset, quantity in quantities:
            result[local_offset:local_offset+local_size] = quantity
        return result
        
    def do_get_attribute(self, reference, name_of_the_attribute, output):
        real_particles,local_offset, local_size, global_size = self.references_to_particles[reference]
        
        try:
            quantity = [local_size, local_offset, getattr(real_particles, name_of_the_attribute)]
        except:
            quantity = [0, 0, None]
        
        quantities = self.mpi_comm.gather(quantity, root = 0)
        
        if self.rank  == 0:
           value = self._merge_results(quantities)
        else:
            value = None
        output.value = dump_and_encode(value)
        return 0
        
    def do_getitem(self, reference_in, pickled_index, is_particle_out, reference_out):
        real_particles, local_offset, local_size, global_size  = self.references_to_particles[reference_in]
        index = decode_and_load(pickled_index)
        self.references_to_particles[self.reference_counter] = self.EmptyReference
        if isinstance(index, int):
            if index >= local_offset and index < local_offset + local_size:
                output_particles = real_particles[index - local_offset]
                self.references_to_particles[self.reference_counter] = ReferencedParticles(
                    output_particles,
                    0,
                    0,
                    0
                )
            is_particle_out.value = True
        elif isinstance(index, slice):
            start, stop, step = index.indices(global_size)
            start -= local_offset
            stop -= local_offset
            start = min(max(0, start), local_size)
            stop = min(max(0, stop), local_size)
            output_particles = real_particles[start:stop]
            input = numpy.zeros(1,  dtype='int64')
            output = numpy.zeros(self.number_of_processes,  dtype='int64')
            
            input[0] = len(output_particles)
                
            self.mpi_comm.Allgather(
                [input, MPI.INTEGER], 
                [output, MPI.INTEGER]
            )
            total_size = 0
            local_offset = 0
            for i,current_size in enumerate(output):
                if i < self.rank:
                    local_offset += current_size
                total_size += current_size
            referenced_particles = ReferencedParticles(
                output_particles,
                local_offset,
                len(output_particles),
                total_size
            )
            self.references_to_particles[self.reference_counter] = referenced_particles
            
            is_particle_out.value = False # check, a slice could still result in 1 particle!!!
        else:
            raise Exception("need to parse index and do smart things here!!")
        
        reference_out.value = self.reference_counter
        self.reference_counter += 1
        return 0
        
    def set_from_generator(self,reference, pickled_generator, pickled_args=None,pickled_kwargs=None):
        generator=decode_and_load(pickled_generator)
        args=()
        if pickled_args is not None:
          args=decode_and_load(pickled_args)
        kwargs={}
        if pickled_kwargs is not None:
          kwargs=decode_and_load(pickled_kwargs)

        real_particles,local_offset, local_size, global_size = self.references_to_particles[reference]
        particles=generator(local_offset,local_offset+local_size,*args,**kwargs)
        keys = real_particles.get_all_keys_in_store()
        real_particles.remove_particles(real_particles[:])
        attributes = particles.get_attribute_names_defined_in_store()
        indices = particles.get_all_indices_in_store()
        values =  particles.get_values_in_store(indices, attributes)
        result = datamodel.Particles()
        result.add_particles_to_store(keys, attributes, values)
        real_particles.add_particles(result)
        return 0

    def select_array(self, reference_in, pickled_function, pickled_attributes, reference_out):
        real_particles, local_offset, local_size, global_size  = self.references_to_particles[reference_in]
        function = decode_and_load(pickled_function)
        attributes = decode_and_load(pickled_attributes)
        
        output_particles = real_particles.select_array(function, attributes)
        
        input = numpy.zeros(1,  dtype='int64')
        output = numpy.zeros(self.number_of_processes,  dtype='int64')
            
        input[0] = len(output_particles)
                
        self.mpi_comm.Allgather(
                [input, MPI.INTEGER], 
                [output, MPI.INTEGER]
        )
        total_size = 0
        local_offset = 0
        for i,current_size in enumerate(output):
            if i < self.rank:
                local_offset += current_size
            total_size += current_size
        referenced_particles = ReferencedParticles(
            output_particles,
            local_offset,
            len(output_particles),
            total_size
        )
        self.references_to_particles[self.reference_counter] = referenced_particles
        reference_out.value = self.reference_counter
        self.reference_counter += 1
        return 0

    def write_set_to_file(self, reference_in, filebase, fileformat, filenames_out):
        real_particles, local_offset, local_size, global_size  = self.references_to_particles[reference_in]
        filename=filebase+"_%6.6i"%self.rank
        write_set_to_file(real_particles, filename, fileformat, 
           local_offset=local_offset, local_size=local_size, global_size=global_size)
        filenames = self.mpi_comm.gather(filename, root = 0)
        if self.rank==0:
          value= filenames
        else:
          value = None
        filenames_out.value = dump_and_encode(value) 
        return 0        

    def read_set_from_file(self, reference_in, filenames_in, fileformat):
        ceil=lambda x,y: (x//y+(x%y>0))
        
        filenames=decode_and_load(filenames_in)
        if reference_in!=0:
          raise Exception("reference_in hould be zero here!")
        real_particles, local_offset, local_size, global_size  = self.references_to_particles[reference_in]
        start=self.rank*ceil(len(filenames),self.number_of_processes)
        end=(self.rank+1)*ceil(len(filenames),self.number_of_processes)
        for i in range(start,end):
          if i< len(filenames):
            p=read_set_from_file(filenames[i],fileformat)
            real_particles.add_particles(p)

        input = numpy.zeros(1,  dtype='int64')
        output = numpy.zeros(self.number_of_processes,  dtype='int64')
            
        input[0] = len(real_particles)
        self.mpi_comm.Allgather(
                [input, MPI.INTEGER], 
                [output, MPI.INTEGER]
        )
        total_size = 0
        local_offset = 0
        for i,current_size in enumerate(output):
            if i < self.rank:
                local_offset += current_size
            total_size += current_size
        referenced_particles = ReferencedParticles(
            real_particles,
            local_offset,
            len(real_particles),
            total_size
        )
        self.references_to_particles[reference_in] = referenced_particles
        return 0

def generate_set_example_function(start,end,*args,**kwargs):
    from amuse.datamodel import Particles
    p=Particles(end-start)
    p.index=list(range(start,end))
    return p

def distributed_king_generator(start,end,*args,**kwargs):
    from amuse.ic.kingmodel import MakeKingModel
    import numpy
    numpy.random.seed(1234567)
    numpy.random.uniform(size=start*6)
    total_number_of_particles=args[0]
    class PartialKingModel(MakeKingModel):
        def makeking(self):
            m,p,v=MakeKingModel.makeking(self)
            m=0.*m+(1.0 / total_number_of_particles)
            return m,p,v
    kwargs['center_model']=False
    kwargs['do_scale']=False
    args=(end-start,)+args[1:]
    return PartialKingModel(*args,**kwargs).result

def select_example_function(x):
    return x > 5
    
class TestDistributedParticles(TestWithMPI):

    def setUp(self):
        if MPI is None or CodeInterface(must_start_worker=False).channel_type != 'mpi':
            self.skip("test needs mpi")
        
    def test1(self):
        x = DistributedParticles(
            size = 3,
            number_of_workers = 1
        )
        self.assertEqual(len(x) , 3)
        x.mass = 10 | units.MSun
        self.assertEqual(x.mass, [10, 10, 10]| units.MSun)
        

    def test2(self):
        x = DistributedParticles(
            size = 3,
            number_of_workers = 1
        )
        self.assertEqual(len(x) , 3)
        x.mass = 10 | units.MSun
        y = x[0:2]
        self.assertEqual(len(y) , 2)
        self.assertEqual(y.mass, [10, 10]| units.MSun)
        

    def test3(self):
        x = DistributedParticles(
            size = 3,
            number_of_workers = 1
        )
        self.assertEqual(len(x) , 3)
        x.mass = [1,2,3]| units.MSun
        y = x[0:2]
        z = x[1:]
        self.assertEqual(len(y) , 2)
        self.assertEqual(y.mass, [1,2]| units.MSun)
        self.assertEqual(len(z) , 2)
        self.assertEqual(z.mass, [2,3]| units.MSun)
        z.mass = [4,5] | units.MSun
        self.assertEqual(y.mass, [1,4]| units.MSun)
        self.assertEqual(z.mass, [4,5]| units.MSun)
        self.assertEqual(x.mass, [1,4,5]| units.MSun)
        
        
    def test4(self):
        x = DistributedParticles(
            size = 3,
            number_of_workers = 1
        )
        x.mass = [1,2,3]| units.MSun
        y = x[1]
        self.assertEqual(y.mass, 2 | units.MSun)
        y.mass = 10 | units.MSun
        self.assertEqual(y.mass, 10 | units.MSun)
        self.assertEqual(x.mass, [1,10,3]| units.MSun)

        
    def test5(self):
        x = DistributedParticles(
            size = 8,
            number_of_workers = 2
        )
        self.assertEqual(len(x) , 8)
        x.mass = [1,2,3,4,5,6,7,8] | units.MSun
        for index in range(len(x)):
            self.assertEqual(x[index].mass, (index+1)| units.MSun)
            
    
    def test6(self):
        x = DistributedParticles(
            size = 9,
            number_of_workers = 2
        )
        self.assertEqual(len(x) , 9)
        x.mass = [1,2,3,4,5,6,7,8,9] | units.MSun
        for index in range(len(x)):
            self.assertEqual(x[index].mass, (index+1)| units.MSun)
        
    def test7(self):
        x = DistributedParticles(
            size = 8,
            number_of_workers = 2
        )
        self.assertEqual(len(x) , 8)
        x.mass = [1,2,3,4,5,6,7,8] | units.MSun
        self.assertEqual(len(x[3:7]), 4)
        x[3:7].mass = [10,11,12,13] | units.MSun
        expected = [1,2,3,10,11,12,13,8]| units.MSun
        for index in range(len(x)):
            self.assertEqual(x[index].mass, expected[index] )
            
    def test8(self):
        x = DistributedParticles(
            size = 8,
            number_of_workers = 2
        )
        self.assertEqual(len(x) , 8)
        x.mass = [1,2,3,4,5,6,7,8] | units.MSun
        self.assertEqual(len(x[3:7]), 4)
        x[3:7].mass = [10,11,12,13] | units.MSun
        
        self.assertEqual(x[2:6].mass, [3,10,11,12]| units.MSun)
            
    def test9(self):
        x = DistributedParticles(
            size = 8,
            number_of_workers = 2
        )
        self.assertEqual(len(x) , 8)
        x.mass = [1,2,3,4,5,6,7,8] 
        self.assertEqual(len(x[3:7]), 4)
        x[3:7].mass = [10,11,12,13] 
        
        self.assertEqual(x[2:6].mass, [3,10,11,12])
        
    def test10(self):
        x = DistributedParticles(
            size = 40,
            number_of_workers = 4
        )
        self.assertEqual(len(x) , 40)
        x.mass = list(range(40)) 
        self.assertEqual(len(x[15:25]), 10)
        self.assertEqual(x[15:25].mass, list(range(15,25)))
        x[15:25].mass = list(range(10))
        self.assertEqual(x[15:25].mass, list(range(10)))
            
    def test11(self):
        from .test_distributed_particles import generate_set_example_function
        y=generate_set_example_function(0,10)
        
        x = DistributedParticles(
            size = 10,
            number_of_workers = 2
        )
        x.set_from_generator(generate_set_example_function)
        self.assertEqual(y.index,x.index)

    def test12(self):
        from .test_distributed_particles import generate_set_example_function
        from .test_distributed_particles import select_example_function
        y=generate_set_example_function(0,10)
        
        x = DistributedParticles(
            size = 10,
            number_of_workers = 2
        )
        x.set_from_generator(generate_set_example_function)
        self.assertEqual(y.index,x.index)        

        highy=y.select_array( select_example_function, ("index",))
        highx=x.select_array( select_example_function, ("index",))
        self.assertEqual(highy.index,highx.index)        
  
    def test13(self):
        x = DistributedParticles(
            size = 8,
            number_of_workers = 2
        )
        x.index= [0,10,10,10,0,0,10,0]
        x.mass = [1, 2, 3, 4,5,6, 7,8] | units.MSun
        highx=x.select_array( select_example_function, ("index",))
        self.assertEqual(highx.mass, [2,3,4,7] | units.MSun)

    def test14(self):
        test_results_path = self.get_path_to_results()
        output_files=[]
        filebase=os.path.join(test_results_path, "test_distributed_sets")
        for i in [0,1]:
          output_file = filebase+"_%6.6i"%i
          output_files.append(output_file)
        
          if os.path.exists(output_file):
            os.remove(output_file)

        x = DistributedParticles(
            size = 8,
            number_of_workers = 2
        )
        x.index= [0,10,10,10,0,0,10,0]
        x.mass = [1, 2, 3, 4,5,6, 7,8] | units.MSun
        files=x.write_set_to_file(filebase,"amuse")
        self.assertEqual(files,output_files)
        for i,f in enumerate(files):
          self.assertTrue( os.path.isfile(f) )
          p=read_set_from_file(f,"amuse")
          self.assertEqual(len(p),4)
          self.assertEqual(p.collection_attributes.global_size,8)          
          self.assertEqual(p.collection_attributes.local_size,4)          

    def test15(self):
        test_results_path = self.get_path_to_results()
        output_files=[]
        filebase=os.path.join(test_results_path, "test_distributed_sets")
        for i in [0,1]:
          output_file = filebase+"_%6.6i"%i
          output_files.append(output_file)
        
          if os.path.exists(output_file):
            os.remove(output_file)

        x = DistributedParticles(
            size = 8,
            number_of_workers = 2
        )
        x.index= [0,10,10,10,0,0,10,0]
        x.mass = [1, 2, 3, 4,5,6, 7,8] | units.MSun
        highx=x.select_array( select_example_function, ("index",))        
        files=highx.write_set_to_file(filebase,"amuse")
        self.assertEqual(files,output_files)
        expected_local_sizes=[3,1]
        for i,f in enumerate(files):
          self.assertTrue( os.path.isfile(f) )
          p=read_set_from_file(f,"amuse")
          self.assertEqual(len(p),expected_local_sizes[i])
          self.assertEqual(p.collection_attributes.global_size,4)          
          self.assertEqual(p.collection_attributes.local_size,expected_local_sizes[i])

    def test16(self):
        test_results_path = self.get_path_to_results()
        filebase=os.path.join(test_results_path, "test_distributed_sets")
        for i in [0,1]:
          output_file = filebase+"_%6.6i"%i
        
          if os.path.exists(output_file):
            os.remove(output_file)

        x = DistributedParticles(
            size = 8,
            number_of_workers = 2
        )
        x.index= [0,10,10,10,0,0,10,0]
        x.mass = [1, 2, 3, 4,5,6, 7,8] | units.MSun
        files=x.write_set_to_file(filebase,"amuse")

        y= DistributedParticles(
            size = 0,
            number_of_workers = 2
        )

        y.read_set_from_file(files,"amuse")
        self.assertEqual(len(y), len(x) )
        self.assertEqual(y.index, x.index )        
        self.assertEqual(y.mass, x.mass )        
        
    def test17(self):
        test_results_path = self.get_path_to_results()
        filebase=os.path.join(test_results_path, "test_distributed_sets")
        for i in [0,1]:
          output_file = filebase+"_%6.6i"%i
        
          if os.path.exists(output_file):
            os.remove(output_file)

        x = DistributedParticles(
            size = 8,
            number_of_workers = 4
        )
        x.index= [0,10,10,10,0,0,10,0]
        x.mass = [1, 2, 3, 4,5,6, 7,8] | units.MSun
        files=x.write_set_to_file(filebase,"amuse")

        y= DistributedParticles(
            size = 0,
            number_of_workers = 2
        )

        y.read_set_from_file(files,"amuse")
        self.assertEqual(len(y), len(x) )
        self.assertEqual(y.index, x.index )        
        self.assertEqual(y.mass, x.mass )        

# number of workers > number of files
# still problematic, because of non-existing attributes if nothing read
    def test18(self):
        test_results_path = self.get_path_to_results()
        filebase=os.path.join(test_results_path, "test_distributed_sets")
        for i in [0,1]:
          output_file = filebase+"_%6.6i"%i
        
          if os.path.exists(output_file):
            os.remove(output_file)

        x = DistributedParticles(
            size = 8,
            number_of_workers = 2
        )
        x.index= [0,10,10,10,0,0,10,0]
        x.mass = [1, 2, 3, 4,5,6, 7,8] | units.MSun
        files=x.write_set_to_file(filebase,"amuse")

        z= DistributedParticles(
            size = 0,
            number_of_workers = 4
        )

        z.read_set_from_file(files,"amuse")
        self.assertEqual(len(x), len(z) )
        self.assertEqual(x.index, z.index )        
        self.assertEqual(x.mass, z.mass )        

    def test19(self):
        from .test_distributed_particles import distributed_king_generator
        from amuse.ic.kingmodel import MakeKingModel
        
        N=100
        W0=7.
        numpy.random.seed(1234567)
        y=MakeKingModel(N,W0,center_model=False).result
        
        x = DistributedParticles(
            size = N,
            number_of_workers = 4
        )
        x.set_from_generator(distributed_king_generator,args=(N,W0))
        self.assertEqual(y.mass,x.mass)
        self.assertEqual(y.x,x.x)
        self.assertEqual(y.y,x.y)
        self.assertEqual(y.z,x.z)
        self.assertEqual(y.vx,x.vx)
        self.assertEqual(y.vy,x.vy)
        self.assertEqual(y.vz,x.vz)

